"""Unit tests for reviewboard.actions.base.BaseAction.

Version Added:
    8.0
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from django.template import Context
from django.utils.safestring import SafeString
from djblets.pagestate.state import PageState

from reviewboard.actions import ActionPlacement, AttachmentPoint
from reviewboard.actions.base import BaseAction
from reviewboard.actions.errors import MissingActionRendererError
from reviewboard.actions.renderers import (ButtonActionRenderer,
                                           DefaultActionRenderer)
from reviewboard.actions.tests.base import (SpecialButtonActionRenderer,
                                            TestAction)
from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.testing import TestCase

if TYPE_CHECKING:
    from djblets.testing.testcases import ExpectedWarning


class BaseActionTests(TestCase):
    """Unit tests for BaseAction.

    Version Added:
        8.0
    """

    def test_init(self) -> None:
        """Testing BaseAction.__init__"""
        class MyAction(BaseAction):
            action_id = 'test-action'

        with self.assertNoWarnings():
            MyAction()

    def test_init_with_deprecations(self) -> None:
        """Testing BaseAction.__init__ with deprecations"""
        class MyAction(BaseAction):
            action_id = 'test-action'
            template_name = 'my-action.html'
            js_view_class = 'My.Action'

            def get_js_view_data(
                self,
                *,
                context: Context,
            ) -> dict:
                return {}

        warning_list: list[ExpectedWarning] = [
            {
                'cls': RemovedInReviewBoard10_0Warning,
                'message': (
                    'MyAction.template_name is deprecated and support will '
                    'be removed in Review Board 10. Please move any custom '
                    'rendering to a reviewboard.actions.renderers.'
                    'BaseActionRenderer subclass instead.'
                ),
            },
            {
                'cls': RemovedInReviewBoard10_0Warning,
                'message': (
                    'MyAction.js_view_class is deprecated and support will '
                    'be removed in Review Board 10. Please move any custom '
                    'rendering to a reviewboard.actions.renderers.'
                    'BaseActionRenderer subclass instead.'
                ),
            },
            {
                'cls': RemovedInReviewBoard10_0Warning,
                'message': (
                    'MyAction.get_js_view_data is deprecated and support will '
                    'be removed in Review Board 10. Please move any custom '
                    'rendering to a reviewboard.actions.renderers.'
                    'BaseActionRenderer subclass instead.'
                ),
            },
        ]

        with self.assertWarnings(warning_list):
            MyAction()

    def test_get_dom_element_id(self) -> None:
        """Testing BaseAction.get_dom_element_id"""
        action = TestAction()

        with self.assertWarns(RemovedInReviewBoard10_0Warning):
            dom_element_id = action.get_dom_element_id()

        # The ID matches what the renderer computes for the placement.
        self.assertEqual(dom_element_id, 'action-review-request-test')

    def test_get_dom_element_id_with_placement_dom_element_id(self) -> None:
        """Testing BaseAction.get_dom_element_id with placement.dom_element_id
        """
        class MyAction(BaseAction):
            action_id = 'placement-test'
            placements = [
                ActionPlacement(attachment=AttachmentPoint.REVIEW_REQUEST,
                                dom_element_id='placement-id'),
            ]

        action = MyAction()

        with self.assertWarns(RemovedInReviewBoard10_0Warning):
            dom_element_id = action.get_dom_element_id()

        self.assertEqual(dom_element_id, 'placement-id')

    def test_get_dom_element_id_with_multiple_placements(self) -> None:
        """Testing BaseAction.get_dom_element_id with multiple placements"""
        class MyAction(BaseAction):
            action_id = 'test-action'
            placements = [
                ActionPlacement(attachment=AttachmentPoint.HEADER),
                ActionPlacement(attachment=AttachmentPoint.REVIEW_REQUEST),
            ]

        action = MyAction()

        # The method can only resolve against the first placement. Renderers
        # compute the ID per-placement instead, which is why this is
        # deprecated. This is here for backwards compatibility.
        with self.assertWarns(RemovedInReviewBoard10_0Warning):
            dom_element_id = action.get_dom_element_id()

        self.assertEqual(dom_element_id, 'action-header-test-action')

    def test_should_register_with_rendered(self) -> None:
        """Testing BaseAction.should_register with rendered"""
        class MyAction(BaseAction):
            action_id = 'test-action'
            label = 'My Label'
            default_renderer_cls = DefaultActionRenderer

        action = MyAction()

        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        MyAction().render_js(request=request,
                             context=context)

        self.assertTrue(action.should_register(request=request,
                                               context=context))

    def test_should_register_with_not_rendered(self) -> None:
        """Testing BaseAction.should_register with not rendered"""
        class MyAction(BaseAction):
            action_id = 'test-action'
            label = 'My Label'
            default_renderer_cls = DefaultActionRenderer

        action = MyAction()

        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        self.assertFalse(action.should_register(request=request,
                                                context=context))

    def test_should_register_with_should_render_true(self) -> None:
        """Testing BaseAction.should_register with should_render() == True"""
        class MyAction(BaseAction):
            action_id = 'test-action'
            label = 'My Label'
            default_renderer_cls = DefaultActionRenderer

            def should_render(
                self,
                *,
                context: Context,
            ) -> bool:
                return True

        action = MyAction()

        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        MyAction().render_js(request=request,
                             context=context)

        self.assertTrue(action.should_register(request=request,
                                               context=context))

    def test_should_register_with_should_render_false(self) -> None:
        """Testing BaseAction.should_register with should_render() == False"""
        class MyAction(BaseAction):
            action_id = 'test-action'
            label = 'My Label'
            default_renderer_cls = DefaultActionRenderer

            def should_render(
                self,
                *,
                context: Context,
            ) -> bool:
                return False

        action = MyAction()

        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        MyAction().render_js(request=request,
                             context=context)

        self.assertFalse(action.should_register(request=request,
                                                context=context))

    def test_render_with_default_renderer(self) -> None:
        """Testing BaseAction.render with default renderer"""
        class MyAction(BaseAction):
            action_id = 'test-action'
            label = 'My Label'
            default_renderer_cls = DefaultActionRenderer

        action = MyAction()

        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        html = action.render(
            request=request,
            context=context,
            placement=action.get_placement('review-request'))

        self.assertIsInstance(html, SafeString)
        self.assertHTMLEqual(
            html,
            """
            <li class="rb-c-actions__action" role="presentation">
             <a href="#"
                id="action-review-request-test-action"
                role="button">
              My Label
             </a>
            </li>
            """)

    def test_render_with_custom_default_renderer(self) -> None:
        """Testing BaseAction.render with custom default_renderer_cls"""
        class MyAction(BaseAction):
            action_id = 'test-action'
            default_renderer_cls = ButtonActionRenderer
            label = 'My Label'

        action = MyAction()

        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        html = action.render(
            request=request,
            context=context,
            placement=action.get_placement('review-request'))

        self.assertIsInstance(html, SafeString)
        self.assertHTMLEqual(
            html,
            """
            <li class="rb-c-actions__action" role="presentation">
             <button aria-label="My Label"
                     class="ink-c-button"
                     id="action-review-request-test-action"
                     type="button">
              <label class="ink-c-button__label">My Label</label>
             </button>
            </li>
            """)

    def test_render_with_renderer(self) -> None:
        """Testing BaseAction.render with provided renderer="""
        class MyAction(BaseAction):
            action_id = 'test-action'
            label = 'My Label'

        action = MyAction()

        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        html = action.render(
            request=request,
            context=context,
            renderer=ButtonActionRenderer,
            placement=action.get_placement('review-request'))

        self.assertIsInstance(html, SafeString)
        self.assertHTMLEqual(
            html,
            """
            <li class="rb-c-actions__action" role="presentation">
             <button aria-label="My Label"
                     class="ink-c-button"
                     id="action-review-request-test-action"
                     type="button">
              <label class="ink-c-button__label">My Label</label>
             </button>
            </li>
            """)

    def test_render_with_no_renderer(self) -> None:
        """Testing BaseAction.render with no renderer"""
        class MyAction(BaseAction):
            action_id = 'test-action'
            label = 'My Label'
            default_renderer_cls = None

        action = MyAction()

        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        message = (
            "A renderer must be explicitly provided when rendering action "
            "<class 'reviewboard.actions.tests.test_base_action."
            "BaseActionTests.test_render_with_no_renderer.<locals>."
            "MyAction'>."
        )

        with self.assertRaisesMessage(MissingActionRendererError, message):
            action.render(request=request,
                          context=context,
                          placement=action.get_placement('review-request'))

    def test_render_with_should_render_false(self) -> None:
        """Testing BaseAction.render with should_render() returning False"""
        class MyAction(BaseAction):
            action_id = 'test-action'
            label = 'My Label'
            default_renderer_cls = ButtonActionRenderer

            def should_render(
                self,
                *,
                context: Context,
            ) -> bool:
                return False

        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        html = MyAction().render(request=request,
                                 context=context)

        self.assertIsInstance(html, SafeString)
        self.assertEqual(html, '')

    def test_render_model_js(self) -> None:
        """Testing BaseAction.render_model_js"""
        class MyAction(BaseAction):
            action_id = 'test-action'
            label = 'My Label'

        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        js = MyAction().render_model_js(request=request,
                                        context=context)

        self.assertIsInstance(js, SafeString)
        self.assertHTMLEqual(
            js,
            """
            page.addAction(new RB.Actions.Action(
                {"id": "test-action",
                 "visible": true,
                 "label": "My Label",
                 "url": "#"},
                { parse: true }
            ));
            """)

    def test_render_js_with_default_renderer(self) -> None:
        """Testing BaseAction.render_js with default renderer"""
        class MyAction(BaseAction):
            action_id = 'test-action'
            label = 'My Label'
            default_renderer_cls = DefaultActionRenderer

        action = MyAction()

        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        js = action.render_js(
            request=request,
            context=context,
            placement=action.get_placement('review-request'))

        self.assertIsInstance(js, SafeString)
        self.assertHTMLEqual(
            js,
            """
            page.addActionView(new RB.Actions.ActionView({
                "attachmentPointID": "review-request",
                el: $('#action-review-request-test-action'),
                model: page.getAction("test-action"),
            }));
            """)

        page_state = PageState.for_request(request)
        self.assertIn('rb-actions-rendered', page_state.extra_data)
        self.assertEqual(
            page_state.extra_data['rb-actions-rendered'],
            {
                action.action_id,
            })

    def test_render_js_with_custom_default_renderer(self) -> None:
        """Testing BaseAction.render_js with custom default_renderer_cls"""
        class MyAction(BaseAction):
            action_id = 'test-action'
            default_renderer_cls = ButtonActionRenderer
            label = 'My Label'

        action = MyAction()

        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        js = action.render_js(
            request=request,
            context=context,
            placement=action.get_placement('review-request'))

        self.assertIsInstance(js, SafeString)
        self.assertHTMLEqual(
            js,
            """
            page.addActionView(new RB.Actions.ButtonActionView({
                "attachmentPointID": "review-request",
                el: $('#action-review-request-test-action'),
                model: page.getAction("test-action"),
            }));
            """)

        page_state = PageState.for_request(request)
        self.assertIn('rb-actions-rendered', page_state.extra_data)
        self.assertEqual(
            page_state.extra_data['rb-actions-rendered'],
            {
                action.action_id,
            })

    def test_render_js_with_renderer(self) -> None:
        """Testing BaseAction.render_js with provided renderer="""
        class MyAction(BaseAction):
            action_id = 'test-action'
            label = 'My Label'

        action = MyAction()

        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        js = action.render_js(
            request=request,
            context=context,
            renderer=SpecialButtonActionRenderer,
            placement=action.get_placement('review-request'))

        self.assertIsInstance(js, SafeString)
        self.assertHTMLEqual(
            js,
            """
            page.addActionView(new SpecialButtonActionView({
                "attachmentPointID": "review-request",
                "label": "~~My Label~~",
                "specialKey": [123, 456],
                el: $('#action-review-request-test-action'),
                model: page.getAction("test-action"),
            }));
            """)

        page_state = PageState.for_request(request)
        self.assertIn('rb-actions-rendered', page_state.extra_data)
        self.assertEqual(
            page_state.extra_data['rb-actions-rendered'],
            {
                action.action_id,
            })

    def test_render_js_with_no_renderer(self) -> None:
        """Testing BaseAction.render_js with no renderer"""
        class MyAction(BaseAction):
            action_id = 'test-action'
            label = 'My Label'
            default_renderer_cls = None

        action = MyAction()

        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        message = (
            "A renderer must be explicitly provided when rendering "
            "JavaScript for action <class 'reviewboard.actions.tests."
            "test_base_action.BaseActionTests.test_render_js_with_no_renderer."
            "<locals>.MyAction'>."
        )

        with self.assertRaisesMessage(MissingActionRendererError, message):
            action.render_js(request=request,
                             context=context,
                             placement=action.get_placement('review-request'))

        page_state = PageState.for_request(request)
        self.assertNotIn('rb-actions-rendered', page_state.extra_data)

    def test_render_js_with_should_render_false(self) -> None:
        """Testing BaseAction.render_js with should_render() returning False"""
        class MyAction(BaseAction):
            action_id = 'test-action'
            label = 'My Label'
            default_renderer_cls = DefaultActionRenderer

            def should_render(
                self,
                *,
                context: Context,
            ) -> bool:
                return False

        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        js = MyAction().render_js(request=request,
                                  context=context)

        self.assertIsInstance(js, SafeString)
        self.assertEqual(js, '')

        page_state = PageState.for_request(request)
        self.assertNotIn('rb-actions-rendered', page_state.extra_data)



"""Management command to reset review request counters on accounts."""

from django.core.management.base import BaseCommand
from django.utils.translation import gettext as _

from reviewboard.accounts.admin import fix_review_counts


class Command(BaseCommand):
    """Management command to reset review request counters on accounts."""

    help = _('Resets all review request-related counters on accounts.')

    def handle(self, **options):
        """Handle the command.

        Args:
            **options (dict, unused):
                Options parsed on the command line. For this command, no
                options are available.
        """
        fix_review_counts()



from __future__ import unicode_literals

import kgb
from djblets.registries.errors import ItemLookupError

from reviewboard.reviews.fields import (BaseReviewRequestField,
                                        BaseReviewRequestFieldSet,
                                        field_registry,
                                        logger)
from reviewboard.testing import TestCase


class MyField(BaseReviewRequestFieldSet):
    field_id = 'test-field'


class BaseReviewRequestFieldSetTests(kgb.SpyAgency, TestCase):
    """Unit tests for BaseReviewRequestFieldSet."""

    def test_is_empty_with_classes(self):
        """Testing BaseReviewRequestFieldSet.is_empty with registered
        field_classes
        """
        class MyFieldset(BaseReviewRequestFieldSet):
            field_classes = [MyField]

        self.assertFalse(MyFieldset.is_empty())

    def test_is_empty_without_classes(self):
        """Testing BaseReviewRequestFieldSet.is_empty without registered
        field_classes
        """
        class MyFieldset(BaseReviewRequestFieldSet):
            pass

        self.assertTrue(MyFieldset.is_empty())

    def test_add_field(self):
        """Testing BaseReviewRequestFieldSet.add_field"""
        class MyFieldset(BaseReviewRequestFieldSet):
            pass

        try:
            MyFieldset.add_field(MyField)

            self.assertEqual(MyFieldset.field_classes, [MyField])
            self.assertIn(MyField, field_registry)
        finally:
            try:
                field_registry.unregister(MyField)
            except Exception:
                pass

    def test_remove_field(self):
        """Testing BaseReviewRequestFieldSet.remove_field"""
        class MyFieldset(BaseReviewRequestFieldSet):
            pass

        try:
            MyFieldset.add_field(MyField)
            MyFieldset.remove_field(MyField)

            self.assertEqual(MyFieldset.field_classes, [])
            self.assertNotIn(MyField, field_registry)
        finally:
            try:
                field_registry.unregister(MyField)
            except Exception:
                pass

    def test_remove_field_with_unregistered(self):
        """Testing BaseReviewRequestFieldSet.remove_field with unregistered
        field
        """
        class MyFieldset(BaseReviewRequestFieldSet):
            pass

        self.spy_on(logger.error)

        with self.assertRaises(ItemLookupError):
            MyFieldset.remove_field(MyField)

        self.assertSpyCalledWith(
            logger.error,
            'Failed to unregister unknown review request field "%s"',
            'test-field')

    def test_should_render_with_visible_fields(self):
        """Testing BaseReviewRequestFieldSet.should_render with visible fields
        """
        class HiddenField(BaseReviewRequestField):
            should_render = False

        class VisibleField(BaseReviewRequestField):
            should_render = True

        class MyFieldset(BaseReviewRequestFieldSet):
            field_classes = [HiddenField, VisibleField]

        user = self.create_user()
        review_request = self.create_review_request(submitter=user)
        fieldset = MyFieldset(review_request_details=review_request)

        self.assertTrue(fieldset.should_render)

    def test_should_render_without_visible_fields(self):
        """Testing BaseReviewRequestFieldSet.should_render without visible
        fields
        """
        class HiddenField(BaseReviewRequestField):
            should_render = False

        class MyFieldset(BaseReviewRequestFieldSet):
            field_classes = [HiddenField]

        user = self.create_user()
        review_request = self.create_review_request(submitter=user)
        fieldset = MyFieldset(review_request_details=review_request)

        self.assertFalse(fieldset.should_render)

    def test_should_render_with_field_error(self):
        """Testing BaseReviewRequestFieldSet.should_render with
        error in Field.should_render
        """
        e = Exception('oh no')

        class HiddenField(BaseReviewRequestField):
            @property
            def should_render(self):
                raise e

        class VisibleField(BaseReviewRequestField):
            should_render = True

        class MyFieldset(BaseReviewRequestFieldSet):
            field_classes = [HiddenField, VisibleField]

        self.spy_on(logger.exception)

        user = self.create_user()
        review_request = self.create_review_request(submitter=user)
        fieldset = MyFieldset(review_request_details=review_request)

        self.assertTrue(fieldset.should_render)
        self.assertSpyCalledWith(
            logger.exception,
            'Failed to call %s.should_render: %s',
            'HiddenField',
            e)

    def test_build_fields(self):
        class MyField1(BaseReviewRequestField):
            pass

        class MyField2(BaseReviewRequestField):
            pass

        class MyField3(BaseReviewRequestField):
            pass

        class MyFieldset(BaseReviewRequestFieldSet):
            def build_fields(self):
                review_request_details = self.review_request_details

                return [
                    MyField1(review_request_details=review_request_details),
                    MyField2(review_request_details=review_request_details),
                    MyField3(review_request_details=review_request_details),
                    MyField2(review_request_details=review_request_details),
                    MyField1(review_request_details=review_request_details),
                ]

        user = self.create_user()
        review_request = self.create_review_request(submitter=user)
        fieldset = MyFieldset(review_request_details=review_request)

        fields = fieldset.build_fields()

        self.assertEqual(len(fields), 5)
        self.assertIsInstance(fields[0], MyField1)
        self.assertIsInstance(fields[1], MyField2)
        self.assertIsInstance(fields[2], MyField3)
        self.assertIsInstance(fields[3], MyField2)
        self.assertIsInstance(fields[4], MyField1)



"""Main application views for Review Board."""

from reviewboard.reviews.views.attachments import (
    DownloadFileAttachmentView,
    ReviewFileAttachmentView,
    ReviewScreenshotView)
from reviewboard.reviews.views.batch import BatchOperationView
from reviewboard.reviews.views.bug_trackers import (BugInfoboxView,
                                                    BugURLRedirectView)
from reviewboard.reviews.views.diff_fragments import (
    CommentDiffFragmentsView,
    ReviewsDiffFragmentView,
    build_diff_comment_fragments)
from reviewboard.reviews.views.diffviewer import ReviewsDiffViewerView
from reviewboard.reviews.views.email import (PreviewBatchEmailView,
                                             PreviewReplyEmailView,
                                             PreviewReviewEmailView,
                                             PreviewReviewRequestEmailView)
from reviewboard.reviews.views.download_diff import (
    DownloadDiffFileView,
    DownloadRawDiffView,
    ReviewsDownloadPatchErrorBundleView)
from reviewboard.reviews.views.mixins import ReviewRequestViewMixin
from reviewboard.reviews.views.new_review_request import NewReviewRequestView
from reviewboard.reviews.views.review_request_detail import \
    ReviewRequestDetailView
from reviewboard.reviews.views.review_request_infobox import \
    ReviewRequestInfoboxView
from reviewboard.reviews.views.review_request_updates import \
    ReviewRequestUpdatesView
from reviewboard.reviews.views.root import RootView


__all__ = [
    'BatchOperationView',
    'BugInfoboxView',
    'BugURLRedirectView',
    'CommentDiffFragmentsView',
    'DownloadDiffFileView',
    'DownloadFileAttachmentView',
    'DownloadRawDiffView',
    'NewReviewRequestView',
    'PreviewBatchEmailView',
    'PreviewReplyEmailView',
    'PreviewReviewEmailView',
    'PreviewReviewRequestEmailView',
    'ReviewFileAttachmentView',
    'ReviewRequestDetailView',
    'ReviewRequestInfoboxView',
    'ReviewRequestUpdatesView',
    'ReviewRequestViewMixin',
    'ReviewScreenshotView',
    'ReviewsDiffFragmentView',
    'ReviewsDiffViewerView',
    'ReviewsDownloadPatchErrorBundleView',
    'RootView',
    'build_diff_comment_fragments',
]

__autodoc_excludes__ = __all__



"""Views for e-mail previews."""

from typing import Any, Dict, Optional

from django.http import HttpRequest
from django.shortcuts import get_object_or_404

from reviewboard.notifications.email.message import (
    prepare_batch_review_request_mail,
    prepare_reply_published_mail,
    prepare_review_published_mail,
    prepare_review_request_mail)
from reviewboard.notifications.email.views import BasePreviewEmailView
from reviewboard.reviews.models import Review
from reviewboard.reviews.views.mixins import ReviewRequestViewMixin


class PreviewBatchEmailView(ReviewRequestViewMixin,
                            BasePreviewEmailView):
    """Display a preview of batch e-mails.

    This can be used to see what an HTML or plain text e-mail will look like
    for a batch publish operation.

    This supports several query parameters to control what data is included in
    the preview:

    changedesc_id (int):
        The PK of the ChangeDescription for displaying a review request update

    reviews_only (int):
        If present, will render as if no review request is being published,
        only reviews and review replies.

    reviews (comma-separated list of int):
        The PKs of reviews to include.

    review_replies (comma-separated list of int):
        The PKs of review replies to include.

    Version Added:
        6.0
    """

    build_email = staticmethod(prepare_batch_review_request_mail)

    def get_email_data(
        self,
        request: HttpRequest,
        *args,
        **kwargs,
    ) -> Dict[str, Any]:
        """Return data used for the e-mail builder.

        The data returned will be passed to :py:attr:`build_email` to handle
        rendering the e-mail.

        This can also return a :py:class:`~django.http.HttpResponse`, which
        is useful for returning errors.

        Args:
            request (django.http.HttpResponse):
                The HTTP response from the client.

            *args (tuple):
                Additional positional arguments passed to the handler.

            **kwargs (dict):
                Additional keyword arguments passed to the handler.

        Returns:
            object:
            The dictionary data to pass as keyword arguments to
            :py:attr:`build_email`, or an instance of
            :py:class:`~django.http.HttpResponse` to immediately return to
            the client.

        Raises:
            django.http.Http404:
                The provided change description did not exist.
        """
        changedesc_id = request.GET.get('changedesc')
        reviews_only = request.GET.get('reviews_only')
        reviews = request.GET.get('reviews', '').split(',')
        review_ids = set(int(pk) for pk in reviews if pk != '')
        replies = request.GET.get('review_replies', '').split(',')
        reply_ids = set(int(pk) for pk in replies if pk != '')

        if changedesc_id:
            changedesc = get_object_or_404(self.review_request.changedescs,
                                           pk=int(changedesc_id))
            user = changedesc.get_user(self.review_request)
        else:
            changedesc = None
            user = self.review_request.submitter

        all_reviews = list(
            Review.objects.filter(pk__in=(review_ids | reply_ids)))

        return {
            'changedesc': changedesc,
            'review_request_changed': not reviews_only,
            'request': request,
            'review_replies': [review for review in all_reviews
                               if review.pk in review_ids],
            'review_request': self.review_request,
            'reviews': [review for review in all_reviews
                        if review.pk in reply_ids],
            'user': user,
        }


class PreviewReviewRequestEmailView(ReviewRequestViewMixin,
                                    BasePreviewEmailView):
    """Display a preview of an e-mail for a review request.

    This can be used to see what an HTML or plain text e-mail will look like
    for a newly-posted review request or an update to a review request.
    """

    build_email = staticmethod(prepare_review_request_mail)

    def get_email_data(
        self,
        request: HttpRequest,
        changedesc_id: Optional[int] = None,
        *args,
        **kwargs,
    ) -> Dict[str, Any]:
        """Return data used for the e-mail builder.

        The data returned will be passed to :py:attr:`build_email` to handle
        rendering the e-mail.

        This can also return a :py:class:`~django.http.HttpResponse`, which
        is useful for returning errors.

        Args:
            request (django.http.HttpResponse):
                The HTTP response from the client.

            changedesc_id (int, optional):
                The ID of a change description used when previewing a
                Review Request Updated e-mail.

            *args (tuple):
                Additional positional arguments passed to the handler.

            **kwargs (dict):
                Additional keyword arguments passed to the handler.

        Returns:
            object:
            The dictionary data to pass as keyword arguments to
            :py:attr:`build_email`, or an instance of
            :py:class:`~django.http.HttpResponse` to immediately return to
            the client.
        """
        close_type = None

        if changedesc_id:
            changedesc = get_object_or_404(self.review_request.changedescs,
                                           pk=changedesc_id)
            user = changedesc.get_user(self.review_request)

            if 'status' in changedesc.fields_changed:
                close_type = changedesc.fields_changed['status']['new'][0]
        else:
            changedesc = None
            user = self.review_request.submitter

        return {
            'user': user,
            'review_request': self.review_request,
            'changedesc': changedesc,
            'close_type': close_type,
        }


class PreviewReviewEmailView(ReviewRequestViewMixin, BasePreviewEmailView):
    """Display a preview of an e-mail for a review.

    This can be used to see what an HTML or plain text e-mail will look like
    for a review.
    """

    build_email = staticmethod(prepare_review_published_mail)

    def get_email_data(
        self,
        request: HttpRequest,
        review_id: int,
        *args,
        **kwargs,
    ) -> Dict[str, Any]:
        """Return data used for the e-mail builder.

        The data returned will be passed to :py:attr:`build_email` to handle
        rendering the e-mail.

        This can also return a :py:class:`~django.http.HttpResponse`, which
        is useful for returning errors.

        Args:
            request (django.http.HttpResponse):
                The HTTP response from the client.

            review_id (int):
                The ID of the review to preview.

            *args (tuple):
                Additional positional arguments passed to the handler.

            **kwargs (dict):
                Additional keyword arguments passed to the handler.

        Returns:
            object:
            The dictionary data to pass as keyword arguments to
            :py:attr:`build_email`, or an instance of
            :py:class:`~django.http.HttpResponse` to immediately return to
            the client.
        """
        review = get_object_or_404(Review,
                                   pk=review_id,
                                   review_request=self.review_request)

        return {
            'user': review.user,
            'review': review,
            'review_request': self.review_request,
            'to_owner_only': False,
            'request': request,
        }


class PreviewReplyEmailView(ReviewRequestViewMixin, BasePreviewEmailView):
    """Display a preview of an e-mail for a reply to a review.

    This can be used to see what an HTML or plain text e-mail will look like
    for a reply to a review.
    """

    build_email = staticmethod(prepare_reply_published_mail)

    def get_email_data(
        self,
        request: HttpRequest,
        review_id: int,
        reply_id: int,
        *args,
        **kwargs,
    ) -> Dict[str, Any]:
        """Return data used for the e-mail builder.

        The data returned will be passed to :py:attr:`build_email` to handle
        rendering the e-mail.

        This can also return a :py:class:`~django.http.HttpResponse`, which
        is useful for returning errors.

        Args:
            request (django.http.HttpResponse):
                The HTTP response from the client.

            review_id (int):
                The ID of the review the reply is for.

            reply_id (int):
                The ID of the reply to preview.

            *args (tuple):
                Additional positional arguments passed to the handler.

            **kwargs (dict):
                Additional keyword arguments passed to the handler.

        Returns:
            object:
            The dictionary data to pass as keyword arguments to
            :py:attr:`build_email`, or an instance of
            :py:class:`~django.http.HttpResponse` to immediately return to
            the client.
        """
        review = get_object_or_404(Review,
                                   pk=review_id,
                                   review_request=self.review_request)
        reply = get_object_or_404(Review, pk=reply_id, base_reply_to=review)

        return {
            'user': reply.user,
            'reply': reply,
            'review': review,
            'review_request': self.review_request,
        }



"""Unit tests for reviewboard.certs.http.urlopen.

Version Added:
    8.1
"""

from __future__ import annotations

import os
from http.client import HTTPConnection
from typing import TYPE_CHECKING
from urllib.request import HTTPHandler, HTTPSHandler, Request

import kgb

from reviewboard.certs.cert import (CertPurpose,
                                    CertificateBundle)
from reviewboard.certs.http import urlopen
from reviewboard.certs.manager import CertificateManager
from reviewboard.certs.tests.testcases import (CaptureSSLMixin,
                                               CertificateTestCase,
                                               MockHTTPResponse,
                                               TEST_CERT_BUNDLE_PEM,
                                               TEST_CLIENT_CERT_PEM,
                                               TEST_CLIENT_KEY_PEM,
                                               TEST_TRUST_CERT_PEM)

if TYPE_CHECKING:
    from reviewboard.site.models import LocalSite


class URLOpenTests(kgb.SpyAgency, CaptureSSLMixin, CertificateTestCase):
    """Unit tests for urlopen.

    Version Added:
        8.1
    """

    def test_with_http(self) -> None:
        """Testing urlopen with http:// URLs"""
        cert_manager = CertificateManager()

        response = urlopen('http://example.com',
                           cert_manager=cert_manager)

        self.assertEqual(response.read(), b'test')

        self.assertSpyCalled(HTTPHandler.http_open)
        self.assertSpyNotCalled(HTTPSHandler.https_open)

    def test_with_https_and_certs_found(self) -> None:
        """Testing urlopen with https:// URLs and certs found in cert manager
        """
        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend
        self._add_cert_state(cert_manager)

        response = urlopen('https://example.com',
                           cert_manager=cert_manager)

        self.assertEqual(response.read(), b'test')

        self.assertSpyNotCalled(HTTPHandler.http_open)
        self.assertSpyCalled(HTTPSHandler.https_open)

        ssl_contexts = self.ssl_contexts
        self.assertEqual(len(ssl_contexts), 1)
        self.assertAttrsEqual(
            ssl_contexts[0],
            {
                'cadatas': [],
                'cafiles': [
                    os.path.join(storage_backend.storage_path, 'certs',
                                 'trust', 'example.com__443.crt'),
                ],
                'capaths': [
                    os.path.join(storage_backend.storage_path, 'cabundles'),
                ],
                'certfiles': [
                    os.path.join(storage_backend.storage_path, 'certs',
                                 'client', 'example.com__443.crt'),
                ],
                'keyfiles': [
                    os.path.join(storage_backend.storage_path, 'certs',
                                 'client', 'example.com__443.key'),
                ],
                'passwords': [
                    None,
                ],
            })

    def test_with_https_and_certs_not_found(self) -> None:
        """Testing urlopen with https:// URLs and certs not found in cert
        manager
        """
        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        response = urlopen('https://other.com',
                           cert_manager=cert_manager)

        self.assertEqual(response.read(), b'test')

        self.assertSpyNotCalled(HTTPHandler.http_open)
        self.assertSpyCalled(HTTPSHandler.https_open)

        ssl_contexts = self.ssl_contexts
        self.assertEqual(len(ssl_contexts), 1)
        self.assertAttrsEqual(
            ssl_contexts[0],
            {
                'cadatas': [],
                'cafiles': [],
                'capaths': [
                    os.path.join(storage_backend.storage_path, 'cabundles'),
                ],
                'certfiles': [],
                'keyfiles': [],
                'passwords': [],
            })

    def test_with_http_to_https_redirect_with_certs(self) -> None:
        """Testing urlopen with http:// to https:// redirect with certs
        found in cert manager
        """
        self.unspy(HTTPConnection.getresponse)
        self.spy_on(
            HTTPConnection.getresponse,
            owner=HTTPConnection,
            op=kgb.SpyOpReturnInOrder([
                MockHTTPResponse(
                    code=302,
                    reason='Found',
                    headers={
                        'location': 'https://example.com',
                    },
                ),
                MockHTTPResponse(
                    data=b'https result',
                ),
            ]))

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend
        self._add_cert_state(cert_manager)

        response = urlopen('http://first.int',
                           cert_manager=cert_manager)

        self.assertEqual(response.read(), b'https result')

        self.assertSpyCalled(HTTPHandler.http_open)
        self.assertSpyCalled(HTTPSHandler.https_open)

        ssl_contexts = self.ssl_contexts
        self.assertEqual(len(ssl_contexts), 1)
        self.assertAttrsEqual(
            ssl_contexts[0],
            {
                'cadatas': [],
                'cafiles': [
                    os.path.join(storage_backend.storage_path, 'certs',
                                 'trust', 'example.com__443.crt'),
                ],
                'capaths': [
                    os.path.join(storage_backend.storage_path, 'cabundles'),
                ],
                'certfiles': [
                    os.path.join(storage_backend.storage_path, 'certs',
                                 'client', 'example.com__443.crt'),
                ],
                'keyfiles': [
                    os.path.join(storage_backend.storage_path, 'certs',
                                 'client', 'example.com__443.key'),
                ],
                'passwords': [
                    None,
                ],
            })

    def test_with_https_to_https_redirect(self) -> None:
        """Testing urlopen with https:// to https:// redirect"""
        self.unspy(HTTPConnection.getresponse)
        self.spy_on(
            HTTPConnection.getresponse,
            owner=HTTPConnection,
            op=kgb.SpyOpReturnInOrder([
                MockHTTPResponse(
                    code=302,
                    reason='Found',
                    headers={
                        'location': 'https://example.com',
                    },
                ),
                MockHTTPResponse(
                    data=b'https result',
                ),
            ]))

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend
        self._add_cert_state(cert_manager)

        response = urlopen('https://first.int',
                           cert_manager=cert_manager)

        self.assertEqual(response.read(), b'https result')

        self.assertSpyNotCalled(HTTPHandler.http_open)
        self.assertSpyCallCount(HTTPSHandler.https_open, 2)

        ssl_contexts = self.ssl_contexts
        self.assertEqual(len(ssl_contexts), 2)
        self.assertAttrsEqual(
            ssl_contexts[0],
            {
                'cadatas': [],
                'cafiles': [],
                'capaths': [
                    os.path.join(storage_backend.storage_path, 'cabundles'),
                ],
                'certfiles': [],
                'keyfiles': [],
                'passwords': [],
            })
        self.assertAttrsEqual(
            ssl_contexts[1],
            {
                'cadatas': [],
                'cafiles': [
                    os.path.join(storage_backend.storage_path, 'certs',
                                 'trust', 'example.com__443.crt'),
                ],
                'capaths': [
                    os.path.join(storage_backend.storage_path, 'cabundles'),
                ],
                'certfiles': [
                    os.path.join(storage_backend.storage_path, 'certs',
                                 'client', 'example.com__443.crt'),
                ],
                'keyfiles': [
                    os.path.join(storage_backend.storage_path, 'certs',
                                 'client', 'example.com__443.key'),
                ],
                'passwords': [
                    None,
                ],
            })

    def test_with_http_to_https_redirect_without_certs(self) -> None:
        """Testing urlopen with http:// to https:// redirect with certs
        not found in cert manager
        """
        self.unspy(HTTPConnection.getresponse)
        self.spy_on(
            HTTPConnection.getresponse,
            owner=HTTPConnection,
            op=kgb.SpyOpReturnInOrder([
                MockHTTPResponse(
                    code=302,
                    reason='Found',
                    headers={
                        'location': 'https://other.com',
                    },
                ),
                MockHTTPResponse(
                    data=b'https result',
                ),
            ]))

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        response = urlopen('http://first.int',
                           cert_manager=cert_manager)

        self.assertEqual(response.read(), b'https result')

        self.assertSpyCalled(HTTPHandler.http_open)
        self.assertSpyCalled(HTTPSHandler.https_open)

        ssl_contexts = self.ssl_contexts
        self.assertEqual(len(ssl_contexts), 1)
        self.assertAttrsEqual(
            ssl_contexts[0],
            {
                'cadatas': [],
                'cafiles': [],
                'capaths': [
                    os.path.join(storage_backend.storage_path, 'cabundles'),
                ],
                'certfiles': [],
                'keyfiles': [],
                'passwords': [],
            })

    def test_with_request(self) -> None:
        """Testing urlopen with Request object"""
        cert_manager = CertificateManager()

        response = urlopen(Request(url='http://example.com'),
                           cert_manager=cert_manager)

        self.assertEqual(response.read(), b'test')

        self.assertSpyCalled(HTTPHandler.http_open)
        self.assertSpyNotCalled(HTTPSHandler.https_open)
        self.assertEqual(self.ssl_contexts, [])

    def test_with_local_site_with_cert_found(self) -> None:
        """Testing urlopen with LocalSite and certs found in cert manager"""
        local_site = self.create_local_site(name='test-site-1')

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend
        self._add_cert_state(cert_manager,
                             local_site=local_site)

        response = urlopen(Request(url='https://example.com'),
                           local_site=local_site)

        self.assertEqual(response.read(), b'test')

        self.assertSpyNotCalled(HTTPHandler.http_open)
        self.assertSpyCalled(HTTPSHandler.https_open)

        base_path = os.path.join(storage_backend.storage_path,
                                 'sites', 'test-site-1',)

        ssl_contexts = self.ssl_contexts
        self.assertEqual(len(ssl_contexts), 1)
        self.assertAttrsEqual(
            ssl_contexts[0],
            {
                'cadatas': [],
                'cafiles': [
                    os.path.join(base_path, 'certs', 'trust',
                                 'example.com__443.crt'),
                ],
                'capaths': [
                    os.path.join(base_path, 'cabundles'),
                ],
                'certfiles': [
                    os.path.join(base_path, 'certs', 'client',
                                 'example.com__443.crt'),
                ],
                'keyfiles': [
                    os.path.join(base_path, 'certs', 'client',
                                 'example.com__443.key'),
                ],
                'passwords': [
                    None,
                ],
            })

    def test_with_local_site_with_cert_not_found(self) -> None:
        """Testing urlopen with LocalSite and certs found in cert manager"""
        local_site = self.create_local_site(name='test-site-1')

        cert_manager = CertificateManager()
        storage_backend = cert_manager.storage_backend

        # Add into the global state.
        self._add_cert_state(cert_manager)

        response = urlopen(Request(url='https://example.com'),
                           cert_manager=cert_manager,
                           local_site=local_site)

        self.assertEqual(response.read(), b'test')

        self.assertSpyNotCalled(HTTPHandler.http_open)
        self.assertSpyCalled(HTTPSHandler.https_open)

        base_path = os.path.join(storage_backend.storage_path,
                                 'sites', 'test-site-1',)

        ssl_contexts = self.ssl_contexts
        self.assertEqual(len(ssl_contexts), 1)
        self.assertAttrsEqual(
            ssl_contexts[0],
            {
                'cadatas': [],
                'cafiles': [],
                'capaths': [
                    os.path.join(base_path, 'cabundles'),
                ],
                'certfiles': [],
                'keyfiles': [],
                'passwords': [],
            })

    def _add_cert_state(
        self,
        cert_manager: CertificateManager,
        *,
        local_site: (LocalSite | None) = None,
    ) -> None:
        """Add testing certificate state to a CertificateManager.

        Args:
            cert_manager (reviewboard.certs.manager.CertificateManager):
                The certificate manager to add to.

            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site to bind the state to.
        """
        cert_manager.add_ca_bundle(
            CertificateBundle(
                bundle_data=TEST_CERT_BUNDLE_PEM,
                name='my-certs',
            ),
            local_site=local_site,
        )
        cert_manager.add_certificate(
            self.create_certificate(
                cert_data=TEST_TRUST_CERT_PEM,
            ),
            local_site=local_site,
        )
        cert_manager.add_certificate(
            self.create_certificate(
                purpose=CertPurpose.CLIENT,
                cert_data=TEST_CLIENT_CERT_PEM,
                key_data=TEST_CLIENT_KEY_PEM,
            ),
            local_site=local_site,
        )



from __future__ import annotations

from django_assert_queries.testing import assert_queries
from django.db.models import Q
from djblets.testing.decorators import add_fixtures
from djblets.webapi.errors import PERMISSION_DENIED

from reviewboard.attachments.models import FileAttachment
from reviewboard.reviews.models import (
    ReviewRequest,
    ReviewRequestDraft)
from reviewboard.reviews.models.review_request import FileAttachmentState
from reviewboard.webapi.resources import resources
from reviewboard.webapi.testing.queries import (
    get_webapi_request_start_equeries,
)
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import (
    draft_file_attachment_item_mimetype,
    draft_file_attachment_list_mimetype)
from reviewboard.webapi.tests.mixins import BasicTestsMetaclass
from reviewboard.webapi.tests.urls import (get_draft_file_attachment_item_url,
                                           get_draft_file_attachment_list_url)


class ResourceListTests(BaseWebAPITestCase, metaclass=BasicTestsMetaclass):
    """Testing the DraftFileAttachmentResource list APIs."""
    fixtures = ['test_users']
    sample_api_url = 'review-requests/<id>/draft/file-attachments/'
    resource = resources.draft_file_attachment

    def compare_item(self, item_rsp, attachment):
        self.assertEqual(item_rsp['id'], attachment.pk)
        self.assertEqual(item_rsp['filename'], attachment.filename)

    def setup_http_not_allowed_item_test(self, user):
        review_request = self.create_review_request(
            submitter=user)

        return get_draft_file_attachment_list_url(review_request)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name,
                             populate_items):
        review_request = self.create_review_request(
            with_local_site=with_local_site,
            submitter=user)

        if populate_items:
            items = [self.create_file_attachment(review_request, draft=True)]
        else:
            items = []

        return (get_draft_file_attachment_list_url(review_request,
                                                   local_site_name),
                draft_file_attachment_list_mimetype,
                items)

    def test_get_with_non_owner_superuser(self):
        """Testing the GET review-requests/<id>/draft/file-attachments/ API
        with non-owner as superuser
        """
        review_request = self.create_review_request(submitter=self.user,
                                                    publish=True)
        attachment = self.create_file_attachment(review_request, draft=True)

        user = self._login_user(admin=True)
        self.assertNotEqual(user, review_request.submitter)

        rsp = self.api_get(
            get_draft_file_attachment_list_url(review_request),
            expected_mimetype=draft_file_attachment_list_mimetype)
        self.assertEqual(rsp['stat'], 'ok')

        attachments = rsp['draft_file_attachments']
        self.assertEqual(len(attachments), 1)
        self.assertEqual(attachments[0]['id'], attachment.pk)

    @add_fixtures(['test_site'])
    def test_get_with_non_owner_local_site_admin(self):
        """Testing the GET review-requests/<id>/draft/file-attachments/ API
        with non-owner as LocalSite admin
        """
        review_request = self.create_review_request(submitter=self.user,
                                                    with_local_site=True,
                                                    publish=True)
        attachment = self.create_file_attachment(review_request, draft=True)

        user = self._login_user(local_site=True, admin=True)
        self.assertNotEqual(user, review_request.submitter)
        self.assertFalse(user.is_superuser)

        rsp = self.api_get(
            get_draft_file_attachment_list_url(review_request,
                                               self.local_site_name),
            expected_mimetype=draft_file_attachment_list_mimetype)
        self.assertEqual(rsp['stat'], 'ok')

        attachments = rsp['draft_file_attachments']
        self.assertEqual(len(attachments), 1)
        self.assertEqual(attachments[0]['id'], attachment.pk)

    #
    # HTTP POST tests
    #

    def setup_basic_post_test(self, user, with_local_site, local_site_name,
                              post_valid_data):
        review_request = self.create_review_request(
            with_local_site=with_local_site,
            submitter=user)

        if post_valid_data:
            post_data = {
                'path': open(self.get_sample_image_filename(), 'rb'),
                'caption': 'New caption',
            }
        else:
            post_data = {}

        return (get_draft_file_attachment_list_url(review_request,
                                                   local_site_name),
                draft_file_attachment_item_mimetype,
                post_data,
                [review_request])

    def check_post_result(self, user, rsp, review_request):
        draft = review_request.get_draft()
        self.assertIsNotNone(draft)

        self.assertIn('draft_file_attachment', rsp)
        item_rsp = rsp['draft_file_attachment']

        attachment = FileAttachment.objects.get(pk=item_rsp['id'])
        self.assertIn(attachment, draft.file_attachments.all())
        self.assertNotIn(attachment, review_request.file_attachments.all())
        self.compare_item(item_rsp, attachment)

    def test_post_with_permission_denied_error(self):
        """Testing the POST review-requests/<id>/draft/file-attachments/ API
        with Permission Denied error
        """
        review_request = self.create_review_request()
        self.assertNotEqual(review_request.submitter, self.user)

        with open(self.get_sample_image_filename(), 'rb') as f:
            rsp = self.api_post(
                get_draft_file_attachment_list_url(review_request),
                {
                    'caption': 'Trophy',
                    'path': f,
                },
                expected_status=403)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], PERMISSION_DENIED.code)


class ResourceItemTests(BaseWebAPITestCase, metaclass=BasicTestsMetaclass):
    """Testing the DraftFileAttachmentResource item APIs."""
    fixtures = ['test_users']
    sample_api_url = 'review-requests/<id>/draft/file-attachments/<id>/'
    resource = resources.draft_file_attachment

    def compare_item(self, item_rsp, attachment):
        self.assertEqual(item_rsp['id'], attachment.pk)
        self.assertEqual(item_rsp['filename'], attachment.filename)

    def setup_http_not_allowed_list_test(self, user):
        review_request = self.create_review_request(
            submitter=user)
        file_attachment = self.create_file_attachment(review_request,
                                                      draft=True)

        return get_draft_file_attachment_item_url(review_request,
                                                  file_attachment.pk)

    #
    # HTTP DELETE tests
    #

    def setup_basic_delete_test(self, user, with_local_site, local_site_name):
        review_request = self.create_review_request(
            with_local_site=with_local_site,
            submitter=user)
        file_attachment = self.create_file_attachment(review_request,
                                                      draft=True)

        return (get_draft_file_attachment_item_url(review_request,
                                                   file_attachment.pk,
                                                   local_site_name),
                [review_request, file_attachment])

    def check_delete_result(self, user, review_request, file_attachment):
        draft = review_request.get_draft()
        self.assertIsNotNone(draft)
        self.assertNotIn(file_attachment,
                         draft.inactive_file_attachments.all())
        self.assertNotIn(file_attachment, draft.file_attachments.all())
        self.assertNotIn(file_attachment,
                         review_request.file_attachments.all())
        self.assertNotIn(file_attachment,
                         review_request.inactive_file_attachments.all())

        with self.assertRaises(FileAttachment.DoesNotExist):
            FileAttachment.objects.get(pk=file_attachment.pk)

    def test_delete_file_with_non_owner_superuser(self):
        """Testing the DELETE review-requests/<id>/draft/file-attachments/<id>/
        API with non-owner as superuser
        """
        review_request = self.create_review_request(submitter=self.user)
        file_attachment = self.create_file_attachment(review_request,
                                                      draft=True)

        user = self._login_user(admin=True)
        self.api_delete(get_draft_file_attachment_item_url(review_request,
                                                           file_attachment.pk))

        self.check_delete_result(user, review_request, file_attachment)

    @add_fixtures(['test_site'])
    def test_delete_file_with_non_owner_local_site_admin(self):
        """Testing the DELETE review-requests/<id>/draft/file-attachments/<id>/
        API with non-owner as LocalSite admin
        """
        review_request = self.create_review_request(submitter=self.user,
                                                    with_local_site=True,
                                                    publish=True)
        file_attachment = self.create_file_attachment(review_request,
                                                      draft=True)

        user = self._login_user(local_site=True, admin=True)
        self.assertNotEqual(user, self.user)

        self.api_delete(get_draft_file_attachment_item_url(
            review_request, file_attachment.pk, self.local_site_name))

        self.check_delete_result(user, review_request, file_attachment)

    def test_delete_file_with_publish(self):
        """Testing the DELETE review-requests/<id>/draft/file-attachments/<id>/
        API with published file attachment
        """
        review_request = self.create_review_request(submitter=self.user,
                                                    target_people=[self.user])
        file_attachment = self.create_file_attachment(review_request,
                                                      draft=True)
        review_request.get_draft().publish()

        self.api_delete(get_draft_file_attachment_item_url(review_request,
                                                           file_attachment.pk))

        draft = review_request.get_draft()
        file_attachment = FileAttachment.objects.get(pk=file_attachment.pk)

        self.assertFalse(file_attachment.inactive_review_request.exists())
        self.assertIsNotNone(draft)
        self.assertIn(file_attachment,
                      draft.inactive_file_attachments.all())
        self.assertNotIn(file_attachment, draft.file_attachments.all())

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name):
        review_request = self.create_review_request(
            with_local_site=with_local_site,
            submitter=user)

        file_attachment = self.create_file_attachment(review_request,
                                                      draft=True)

        return (get_draft_file_attachment_item_url(review_request,
                                                   file_attachment.pk,
                                                   local_site_name),
                draft_file_attachment_item_mimetype,
                file_attachment)

    def test_get_with_caption(self) -> None:
        """Testing the GET review-requests/<id>/draft/file-attachments/<id>/
        with a draft caption
        """
        review_request = self.create_review_request(submitter=self.user)
        file_attachment = self.create_file_attachment(
            review_request,
            draft=True,
            caption='Published caption',
            draft_caption='Draft caption')

        rsp = self.api_get(
            get_draft_file_attachment_item_url(review_request,
                                               file_attachment.pk),
            expected_mimetype=draft_file_attachment_item_mimetype)
        item_rsp = rsp['draft_file_attachment']

        file_attachment.refresh_from_db()

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(item_rsp['id'], file_attachment.pk)
        self.assertEqual(item_rsp['caption'], 'Draft caption')
        self.assertEqual(file_attachment.draft_caption, 'Draft caption')
        self.assertEqual(file_attachment.caption, 'Published caption')

    def test_get_with_empty_caption(self) -> None:
        """Testing the GET review-requests/<id>/draft/file-attachments/<id>/
        with an empty draft caption
        """
        review_request = self.create_review_request(submitter=self.user)
        file_attachment = self.create_file_attachment(
            review_request,
            draft=True,
            caption='Published caption',
            draft_caption='')

        rsp = self.api_get(
            get_draft_file_attachment_item_url(review_request,
                                               file_attachment.pk),
            expected_mimetype=draft_file_attachment_item_mimetype)
        item_rsp = rsp['draft_file_attachment']

        file_attachment.refresh_from_db()

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(item_rsp['id'], file_attachment.pk)
        self.assertEqual(item_rsp['caption'], '')
        self.assertEqual(file_attachment.draft_caption, '')
        self.assertEqual(file_attachment.caption, 'Published caption')

    #
    # HTTP PUT tests
    #

    def setup_basic_put_test(self, user, with_local_site, local_site_name,
                             put_valid_data):
        review_request = self.create_review_request(
            with_local_site=with_local_site,
            submitter=user)
        file_attachment = self.create_file_attachment(review_request)

        return (get_draft_file_attachment_item_url(review_request,
                                                   file_attachment.pk,
                                                   local_site_name),
                draft_file_attachment_item_mimetype,
                {'caption': 'My new caption'},
                file_attachment,
                [])

    def check_put_result(self, user, item_rsp, file_attachment):
        file_attachment = FileAttachment.objects.get(pk=file_attachment.pk)
        self.assertEqual(item_rsp['id'], file_attachment.pk)
        self.assertEqual(item_rsp['caption'], 'My new caption')
        self.assertEqual(file_attachment.draft_caption, 'My new caption')

    def test_put_with_empty_caption(self) -> None:
        """Testing the PUT review-requests/<id>/draft/file-attachments/<id>/
        with an empty caption
        """
        review_request = self.create_review_request(submitter=self.user)
        file_attachment = self.create_file_attachment(review_request)

        rsp = self.api_put(
            get_draft_file_attachment_item_url(review_request,
                                               file_attachment.pk),
            {
                'caption': '',
            },
            expected_mimetype=draft_file_attachment_item_mimetype)
        item_rsp = rsp['draft_file_attachment']

        file_attachment.refresh_from_db()

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(item_rsp['id'], file_attachment.pk)
        self.assertEqual(item_rsp['caption'], '')
        self.assertEqual(file_attachment.draft_caption, '')
        self.assertEqual(file_attachment.caption, 'My Caption')

    def test_put_with_non_owner_superuser(self):
        """Testing the PUT review-requests/<id>/draft/file-attachments/<id>/
        API with non-owner as superuser
        """
        review_request = self.create_review_request(submitter=self.user)
        file_attachment = self.create_file_attachment(review_request)

        user = self._login_user(admin=True)
        self.assertNotEqual(user, self.user)

        rsp = self.api_put(
            get_draft_file_attachment_item_url(review_request,
                                               file_attachment.pk),
            {
                'caption': 'My new caption',
            },
            expected_mimetype=draft_file_attachment_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.check_put_result(user, rsp['draft_file_attachment'],
                              file_attachment)

    @add_fixtures(['test_site'])
    def test_put_file_with_non_owner_local_site_admin(self):
        """Testing the PUT review-requests/<id>/draft/file-attachments/<id>/
        API with non-owner as LocalSite admin
        """
        review_request = self.create_review_request(submitter=self.user,
                                                    with_local_site=True,
                                                    publish=True)
        file_attachment = self.create_file_attachment(review_request,
                                                      draft=True)

        user = self._login_user(local_site=True, admin=True)
        self.assertNotEqual(user, self.user)

        rsp = self.api_put(
            get_draft_file_attachment_item_url(review_request,
                                               file_attachment.pk,
                                               self.local_site_name),
            {
                'caption': 'My new caption',
            },
            expected_mimetype=draft_file_attachment_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.check_put_result(user, rsp['draft_file_attachment'],
                              file_attachment)

    def test_put_with_pending_deletion_true(self) -> None:
        """Testing the PUT review-requests/<id>/draft/file-attachments/<id>/
        with setting pending_deletion to True
        """
        review_request = self.create_review_request(submitter=self.user)
        file_attachment = self.create_file_attachment(review_request)

        rsp = self.api_put(
            get_draft_file_attachment_item_url(review_request,
                                               file_attachment.pk),
            {
                'pending_deletion': True,
            },
            expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], 105)
        self.assertEqual(
            rsp['fields'],
            {
                'pending_deletion': 'This can only be set to false to undo '
                                    'the pending deletion of a published file '
                                    'attachment. You cannot set this to true.',
            })

    def test_put_with_caption(self) -> None:
        """Testing the PUT review-requests/<id>/draft/file-attachments/<id>/
        with setting a new caption and when a draft already exists
        """
        user = self.user
        assert user is not None

        review_request = self.create_review_request(submitter=user)
        file_attachment = self.create_file_attachment(review_request)
        review_request_draft = self.create_review_request_draft(review_request)

        equeries = get_webapi_request_start_equeries(user=user)
        equeries += [
            {
                'model': ReviewRequest,
                'select_related': {'submitter', 'repository'},
                'where': (Q(local_site=None) &
                          Q(pk=str(review_request.pk))),
            },
            {
                'model': ReviewRequestDraft,
                'select_related': {'review_request'},
                'where': Q(review_request=review_request),
            },
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments':
                        'LEFT OUTER JOIN',
                    'reviews_reviewrequestdraft_file_attachments':
                        'LEFT OUTER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 2,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_file_attachments',
                    'reviews_reviewrequestdraft_file_attachments'
                },
                'where': (
                    Q(Q(Q(review_request=review_request) |
                        Q(drafts=review_request_draft)) &
                      Q(added_in_filediff__isnull=True) &
                      Q(repository__isnull=True) &
                      Q(user__isnull=True)) &
                    Q(pk=str(file_attachment.pk))
                ),
            },
            {
                'extra': {
                    '_prefetch_related_val_fileattachment_id': (
                        '"reviews_reviewrequest_file_attachments".'
                        '"fileattachment_id"',
                        [],
                    ),
                },
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(file_attachments__in=[file_attachment]),
            },
            {
                'extra': {
                    '_prefetch_related_val_fileattachment_id': (
                        '"reviews_reviewrequestdraft_file_attachments".'
                        '"fileattachment_id"',
                        [],
                    ),
                },
                'join_types': {
                    'reviews_reviewrequestdraft_file_attachments':
                        'INNER JOIN',
                },
                'model': ReviewRequestDraft,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequestdraft',
                    'reviews_reviewrequestdraft_file_attachments',
                },
                'where': Q(file_attachments__in=[file_attachment]),
            },
            {
                'model': ReviewRequestDraft,
                'where': Q(review_request=review_request),
            },
            {
                'model': FileAttachment,
                'type': 'UPDATE',
                'where': Q(pk=file_attachment.pk),
            },
            {
                'model': ReviewRequestDraft,
                'type': 'UPDATE',
                'where': Q(pk=review_request_draft.pk)
            },
            {
                'join_types': {
                    'reviews_reviewrequestdraft_file_attachments':
                        'INNER JOIN',
                },
                'model': ReviewRequestDraft,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequestdraft_file_attachments',
                    'reviews_reviewrequestdraft',
                },
                'where': Q(file_attachments__id=file_attachment.pk),
            },
            {
                'model': ReviewRequest,
                'where': Q(id=review_request.pk),
            },
        ]

        # The purpose of this test is to see what queries are being executed,
        # to compare against the queries executed during ``pending_deletion``
        # updates.
        with assert_queries(equeries):
            rsp = self.api_put(
                get_draft_file_attachment_item_url(review_request,
                                                   file_attachment.pk),
                {
                    'caption': 'Updated caption',
                },
                expected_mimetype=draft_file_attachment_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')

    def test_put_with_pending_deletion_false(self) -> None:
        """Testing the PUT review-requests/<id>/draft/file-attachments/<id>/
        with setting pending_deletion to False for a file attachment that
        is currently pending deletion
        """
        user = self.user
        assert user is not None

        review_request = self.create_review_request(submitter=user)
        file_attachment = self.create_file_attachment(review_request)
        review_request_draft = self.create_review_request_draft(review_request)

        # "Delete" the file attachment.
        review_request_draft.inactive_file_attachments.add(file_attachment)
        review_request_draft.file_attachments.remove(file_attachment)

        self.assertEqual(
            review_request.get_file_attachment_state(file_attachment),
            FileAttachmentState.PENDING_DELETION)

        del review_request._file_attachments_data

        equeries = get_webapi_request_start_equeries(user=user)
        equeries += [
            {
                'model': ReviewRequest,
                'select_related': {'submitter', 'repository'},
                'where': (Q(local_site=None) &
                          Q(pk=str(review_request.pk))),
            },
            {
                'model': ReviewRequestDraft,
                'select_related': {'review_request'},
                'where': Q(review_request=review_request),
            },
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments':
                        'LEFT OUTER JOIN',
                    'reviews_reviewrequestdraft_inactive_file_attachments':
                        'LEFT OUTER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 2,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_file_attachments',
                    'reviews_reviewrequestdraft_inactive_file_attachments',
                },
                'where': (
                    Q(Q(Q(review_request=review_request) |
                        Q(inactive_drafts=review_request_draft)) &
                      Q(added_in_filediff__isnull=True) &
                      Q(repository__isnull=True) &
                      Q(user__isnull=True)) &
                    Q(pk=str(file_attachment.pk))
                ),
            },
            {
                'extra': {
                    '_prefetch_related_val_fileattachment_id': (
                        '"reviews_reviewrequest_file_attachments".'
                        '"fileattachment_id"',
                        [],
                    ),
                },
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(file_attachments__in=[file_attachment]),
            },
            {
                'extra': {
                    '_prefetch_related_val_fileattachment_id': (
                        '"reviews_reviewrequestdraft_inactive_'
                        'file_attachments"."fileattachment_id"',
                        [],
                    ),
                },
                'join_types': {
                    'reviews_reviewrequestdraft_inactive_file_attachments':
                        'INNER JOIN',
                },
                'model': ReviewRequestDraft,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequestdraft',
                    'reviews_reviewrequestdraft_inactive_file_attachments',
                },
                'where': Q(inactive_file_attachments__in=[file_attachment]),
            },
            {
                'model': ReviewRequestDraft,
                'where': Q(review_request=review_request),
            },
            {
                'model': ReviewRequest,
                'where': Q(id=review_request.pk),
            },
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(review_request__id=review_request.pk),
            },
            {
                'model': ReviewRequestDraft,
                'where': Q(review_request=review_request),
            },
            {
                'join_types': {
                    'reviews_reviewrequestdraft_inactive_file_attachments':
                        'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequestdraft_inactive_file_attachments',
                },
                'where': Q(inactive_drafts__id=review_request_draft.pk),
            },
            {
                'model': FileAttachment,
                'where': Q(
                    attachment_history=file_attachment.attachment_history_id),
                'values_select': ('pk',),
            },
            {
                'model': ReviewRequestDraft.inactive_file_attachments.through,
                'where': (Q(reviewrequestdraft=(review_request_draft.pk,)) &
                          Q(fileattachment__in={file_attachment.pk})),
            },
            {
                'model': ReviewRequestDraft.inactive_file_attachments.through,
                'type': 'DELETE',
                'where': Q(id__in=[file_attachment.pk]),
            },
            {
                'model': ReviewRequestDraft,
                'type': 'UPDATE',
                'where': Q(pk=review_request_draft.pk),
            },
            {
                'model': ReviewRequestDraft,
                'limit': 1,
                'values_select': ('inactive_file_attachments_count',),
                'where': Q(pk=review_request_draft.pk),
            },
            {
                'model': ReviewRequestDraft.file_attachments.through,
                'values_select': ('fileattachment',),
                'where': (Q(fileattachment__in={file_attachment.pk}) &
                          Q(reviewrequestdraft=review_request_draft.pk)),
            },
            {
                'model': ReviewRequestDraft.file_attachments.through,
                'type': 'INSERT',
            },
            {
                'model': ReviewRequestDraft,
                'type': 'UPDATE',
                'where': Q(pk=review_request_draft.pk),
            },
            {
                'model': ReviewRequestDraft,
                'limit': 1,
                'values_select': ('file_attachments_count',),
                'where': Q(pk=review_request_draft.pk),
            },
            {
                'model': ReviewRequestDraft,
                'type': 'UPDATE',
                'where': Q(pk=review_request_draft.pk)
            },
            {
                'join_types': {
                    'reviews_reviewrequestdraft_file_attachments':
                        'INNER JOIN',
                },
                'model': ReviewRequestDraft,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequestdraft_file_attachments',
                    'reviews_reviewrequestdraft',
                },
                'where': Q(file_attachments__id=file_attachment.pk),
            },
            {
                'model': ReviewRequest,
                'where': Q(id=review_request.pk),
            },
        ]

        with assert_queries(equeries):
            rsp = self.api_put(
                get_draft_file_attachment_item_url(review_request,
                                                   file_attachment.pk),
                {
                    'pending_deletion': False,
                },
                expected_mimetype=draft_file_attachment_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(
            review_request.get_file_attachment_state(file_attachment),
            FileAttachmentState.PUBLISHED)

    def test_put_with_pending_deletion_false_with_history(self) -> None:
        """Testing the PUT review-requests/<id>/draft/file-attachments/<id>/
        with setting pending_deletion to False for a file attachment that
        is currently pending deletion and has a history of revisions
        """
        user = self.user
        assert user is not None

        # Test with a lot of data, to help catch performance issues.
        review_request = self.create_review_request(submitter=user)
        file_attachment = self.create_file_attachment(review_request)
        file_attachment_2 = self.create_file_attachment(
            review_request,
            attachment_history=file_attachment.attachment_history,
            attachment_revision=file_attachment.attachment_revision + 1)
        file_attachment_3 = self.create_file_attachment(
            review_request,
            attachment_history=file_attachment.attachment_history,
            attachment_revision=file_attachment.attachment_revision + 2)
        file_attachment_4 = self.create_file_attachment(
            review_request,
            attachment_history=file_attachment.attachment_history,
            attachment_revision=file_attachment.attachment_revision + 3)
        file_attachment_5 = self.create_file_attachment(
            review_request,
            attachment_history=file_attachment.attachment_history,
            attachment_revision=file_attachment.attachment_revision + 4)
        file_attachment_6 = self.create_file_attachment(
            review_request,
            attachment_history=file_attachment.attachment_history,
            attachment_revision=file_attachment.attachment_revision + 5)
        file_attachment_7 = self.create_file_attachment(
            review_request,
            attachment_history=file_attachment.attachment_history,
            attachment_revision=file_attachment.attachment_revision + 6)
        review_request_draft = self.create_review_request_draft(review_request)

        # "Delete" the file attachment and all revisions of it.
        review_request_draft.inactive_file_attachments.add(
            file_attachment,
            file_attachment_2,
            file_attachment_3,
            file_attachment_4,
            file_attachment_5,
            file_attachment_6,
            file_attachment_7)
        review_request_draft.file_attachments.remove(
            file_attachment,
            file_attachment_2,
            file_attachment_3,
            file_attachment_4,
            file_attachment_5,
            file_attachment_6,
            file_attachment_7)

        self.assertEqual(
            review_request.get_file_attachment_state(file_attachment),
            FileAttachmentState.PENDING_DELETION)
        self.assertEqual(
            review_request.get_file_attachment_state(file_attachment_2),
            FileAttachmentState.PENDING_DELETION)
        self.assertEqual(
            review_request.get_file_attachment_state(file_attachment_3),
            FileAttachmentState.PENDING_DELETION)
        self.assertEqual(
            review_request.get_file_attachment_state(file_attachment_4),
            FileAttachmentState.PENDING_DELETION)
        self.assertEqual(
            review_request.get_file_attachment_state(file_attachment_5),
            FileAttachmentState.PENDING_DELETION)

        del review_request._file_attachments_data

        equeries = get_webapi_request_start_equeries(user=user)
        equeries += [
            {
                'model': ReviewRequest,
                'select_related': {'submitter', 'repository'},
                'where': (Q(local_site=None) &
                          Q(pk=str(review_request.pk))),
            },
            {
                'model': ReviewRequestDraft,
                'select_related': {'review_request'},
                'where': Q(review_request=review_request),
            },
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments':
                        'LEFT OUTER JOIN',
                    'reviews_reviewrequestdraft_inactive_file_attachments':
                        'LEFT OUTER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 2,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_file_attachments',
                    'reviews_reviewrequestdraft_inactive_file_attachments',
                },
                'where': (
                    Q(Q(Q(review_request=review_request) |
                        Q(inactive_drafts=review_request_draft)) &
                      Q(added_in_filediff__isnull=True) &
                      Q(repository__isnull=True) &
                      Q(user__isnull=True)) &
                    Q(pk=str(file_attachment.pk))
                ),
            },
            {
                'extra': {
                    '_prefetch_related_val_fileattachment_id': (
                        '"reviews_reviewrequest_file_attachments".'
                        '"fileattachment_id"',
                        [],
                    ),
                },
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(file_attachments__in=[file_attachment]),
            },
            {
                'extra': {
                    '_prefetch_related_val_fileattachment_id': (
                        '"reviews_reviewrequestdraft_inactive_file_'
                        'attachments"."fileattachment_id"',
                        [],
                    ),
                },
                'join_types': {
                    'reviews_reviewrequestdraft_inactive_file_attachments':
                        'INNER JOIN',
                },
                'model': ReviewRequestDraft,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequestdraft',
                    'reviews_reviewrequestdraft_inactive_file_attachments',
                },
                'where': Q(inactive_file_attachments__in=[file_attachment]),
            },
            {
                'model': ReviewRequestDraft,
                'where': Q(review_request=review_request),
            },
            {
                'model': ReviewRequest,
                'where': Q(id=review_request.pk),
            },
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(review_request__id=review_request.pk),
            },
            {
                'model': ReviewRequestDraft,
                'where': Q(review_request=review_request),
            },
            {
                'join_types': {
                    'reviews_reviewrequestdraft_inactive_file_attachments':
                        'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequestdraft_inactive_file_attachments',
                },
                'where': Q(inactive_drafts__id=review_request_draft.pk),
            },
            {
                'model': FileAttachment,
                'where': Q(
                    attachment_history=file_attachment.attachment_history_id),
                'values_select': ('pk',),
            },
            {
                'model': ReviewRequestDraft.inactive_file_attachments.through,
                'where': (Q(reviewrequestdraft=(review_request_draft.pk,)) &
                          Q(fileattachment__in={
                              file_attachment.pk,
                              file_attachment_2.pk,
                              file_attachment_3.pk,
                              file_attachment_4.pk,
                              file_attachment_5.pk,
                              file_attachment_6.pk,
                              file_attachment_7.pk,
                          })),
            },
            {
                'model': ReviewRequestDraft.inactive_file_attachments.through,
                'type': 'DELETE',
                'where': Q(id__in=[
                    file_attachment_7.pk,
                    file_attachment_6.pk,
                    file_attachment_5.pk,
                    file_attachment_4.pk,
                    file_attachment_3.pk,
                    file_attachment_2.pk,
                    file_attachment.pk,
                ]),
            },
            {
                'model': ReviewRequestDraft,
                'type': 'UPDATE',
                'where': Q(pk=review_request_draft.pk),
            },
            {
                'model': ReviewRequestDraft,
                'limit': 1,
                'values_select': ('inactive_file_attachments_count',),
                'where': Q(pk=review_request_draft.pk),
            },
            {
                'model': ReviewRequestDraft.file_attachments.through,
                'values_select': ('fileattachment',),
                'where': (
                    Q(fileattachment__in={
                        file_attachment.pk,
                        file_attachment_2.pk,
                        file_attachment_3.pk,
                        file_attachment_4.pk,
                        file_attachment_5.pk,
                        file_attachment_6.pk,
                        file_attachment_7.pk,
                    }) &
                    Q(reviewrequestdraft=review_request_draft.pk)),
            },
            {
                'model': ReviewRequestDraft.file_attachments.through,
                'type': 'INSERT',
            },
            {
                'model': ReviewRequestDraft,
                'type': 'UPDATE',
                'where': Q(pk=review_request_draft.pk),
            },
            {
                'model': ReviewRequestDraft,
                'limit': 1,
                'values_select': ('file_attachments_count',),
                'where': Q(pk=review_request_draft.pk),
            },
            {
                'model': ReviewRequestDraft,
                'type': 'UPDATE',
                'where': Q(pk=review_request_draft.pk)
            },
            {
                'join_types': {
                    'reviews_reviewrequestdraft_file_attachments':
                        'INNER JOIN',
                },
                'model': ReviewRequestDraft,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequestdraft',
                    'reviews_reviewrequestdraft_file_attachments',
                },
                'where': Q(file_attachments__id=file_attachment.pk),
            },
            {
                'model': ReviewRequest,
                'where': Q(id=review_request.pk),
            },
        ]

        with assert_queries(equeries):
            rsp = self.api_put(
                get_draft_file_attachment_item_url(review_request,
                                                   file_attachment.pk),
                {
                    'pending_deletion': False,
                },
                expected_mimetype=draft_file_attachment_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(
            review_request.get_file_attachment_state(file_attachment),
            FileAttachmentState.PUBLISHED)
        self.assertEqual(
            review_request.get_file_attachment_state(file_attachment_2),
            FileAttachmentState.PUBLISHED)

    def test_put_with_pending_deletion_false_for_invalid(self) -> None:
        """Testing the PUT review-requests/<id>/draft/file-attachments/<id>/
        with setting pending_deletion to False for a file attachment that isn't
        currently pending deletion
        """
        user = self.user
        assert user is not None

        review_request = self.create_review_request(submitter=user)
        file_attachment = self.create_file_attachment(review_request)
        review_request_draft = self.create_review_request_draft(review_request)

        equeries = get_webapi_request_start_equeries(user=user)
        equeries += [
            {
                'model': ReviewRequest,
                'select_related': {'submitter', 'repository'},
                'where': (Q(local_site=None) &
                          Q(pk=str(review_request.pk))),
            },
            {
                'model': ReviewRequestDraft,
                'select_related': {'review_request'},
                'where': Q(review_request=review_request),
            },
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments':
                        'LEFT OUTER JOIN',
                    'reviews_reviewrequestdraft_file_attachments':
                        'LEFT OUTER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 2,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_file_attachments',
                    'reviews_reviewrequestdraft_file_attachments'
                },
                'where': (
                    Q(Q(Q(review_request=review_request) |
                        Q(drafts=review_request_draft)) &
                      Q(added_in_filediff__isnull=True) &
                      Q(repository__isnull=True) &
                      Q(user__isnull=True)) &
                    Q(pk=str(review_request.pk))
                ),
            },
            {
                'extra': {
                    '_prefetch_related_val_fileattachment_id': (
                        '"reviews_reviewrequest_file_attachments".'
                        '"fileattachment_id"',
                        [],
                    ),
                },
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(file_attachments__in=[file_attachment]),
            },
            {
                'extra': {
                    '_prefetch_related_val_fileattachment_id': (
                        '"reviews_reviewrequestdraft_file_attachments".'
                        '"fileattachment_id"',
                        [],
                    ),
                },
                'join_types': {
                    'reviews_reviewrequestdraft_file_attachments':
                        'INNER JOIN',
                },
                'model': ReviewRequestDraft,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequestdraft',
                    'reviews_reviewrequestdraft_file_attachments',
                },
                'where': Q(file_attachments__in=[file_attachment]),
            },
            {
                'model': ReviewRequestDraft,
                'where': Q(review_request=review_request),
            },
            {
                'model': ReviewRequest,
                'tables': {
                    'reviews_reviewrequest',
                },
                'where': Q(id=review_request_draft.pk),
            },
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(review_request__id=review_request.pk),
            },
            {
                'model': ReviewRequestDraft,
                'limit': 21,
                'where': Q(review_request=review_request),
            },
            {
                'join_types': {
                    'reviews_reviewrequestdraft_file_attachments':
                        'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequestdraft_file_attachments',
                },
                'where': Q(drafts__id=review_request_draft.pk),
            },
        ]

        with assert_queries(equeries):
            rsp = self.api_put(
                get_draft_file_attachment_item_url(review_request,
                                                   file_attachment.pk),
                {
                    'pending_deletion': False,
                },
                expected_status=400)

        self.assertEqual(
            rsp,
            {
                'err': {
                    'code': 105,
                    'msg': 'One or more fields had errors',
                    'type': 'request-field-error',
                },
                'fields': {
                    'pending_deletion': (
                        'This can only be used to undo the pending '
                        'deletion of a file attachment. This file '
                        'attachment is not currently pending '
                        'deletion.'
                    ),
                },
                'stat': 'fail',
            })



"""Representations of remote repositories.

Version Added:
    6.0
    This replaces the old :py:mod:`reviewboard.hostingsvcs.repository` module.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typelets.json import JSONDict

    from reviewboard.hostingsvcs.base import BaseHostingService
    from reviewboard.hostingsvcs.models import HostingServiceAccount


class RemoteRepository:
    """A representation of a remote repository.

    This is used to represent the configuration for a repository that already
    exists on the hosting service. It does not necessarily match a repository
    configured on Review Board, but can be used to create one.

    Version Changed:
        6.0:
        * Moved from :py:mod:`reviewboard.hostingsvcs.repository` to
          :py:mod:`reviewboard.hostingsvcs.base.repository`.
    """

    ######################
    # Instance variables #
    ######################

    #: Extra data about the repository.
    #:
    #: Type:
    #:     dict
    extra_data: JSONDict

    #: The hosting service that owns the repository.
    #:
    #: Type:
    #:     reviewboard.hostingsvcs.base.hosting_service.BaseHostingService
    hosting_service: BaseHostingService

    #: The account used for the hosting service.
    #:
    #: Type:
    #:     reviewboard.hostingsvcs.models.HostingServiceAccount
    hosting_service_account: HostingServiceAccount

    #: The service-specific ID of the repository.
    #:
    #: Type:
    #:     str
    id: str

    #: The mirror (alternate) path of the repository.
    #:
    #: Type:
    #:     str
    mirror_path: str | None

    #: The displayed name of the repository.
    #:
    #: Type:
    #:     str
    name: str

    #: The identifier of the owner of the repository.
    #:
    #: Type:
    #:     str
    owner: str

    #: The repository path.
    #:
    #: Type:
    #:     str
    path: str

    #: The service-specific identifier for the type of repository.
    #:
    #: Type:
    #:     str
    scm_type: str

    def __init__(
        self,
        hosting_service: BaseHostingService,
        repository_id: str,
        name: str,
        owner: str,
        scm_type: str,
        path: str,
        mirror_path: (str | None) = None,
        extra_data: (JSONDict | None) = None,
    ) -> None:
        """Initialize the remote repository representation.

        Args:
            hosting_service (reviewboard.hostingsvcs.base.hosting_service.
                             BaseHostingService):
                The hosting service that owns the repository.

            repository_id (str):
                The service-specific identifier for the type of repository.

            name (str):
                The displayed name of the repository.

            owner (str):
                The identifier of the owner of the repository.

            scm_type (str):
                The service-specific identifier for the type of repository.

            path (str):
                The repository path.

            mirror_path (str, optional):
                The mirror (alternate) path of the repository.

            extra_data (dict, optional):
                Extra data about the repository.
        """
        self.hosting_service = hosting_service
        self.hosting_service_account = hosting_service.account
        self.id = repository_id
        self.name = name
        self.owner = owner
        self.scm_type = scm_type
        self.path = path
        self.mirror_path = mirror_path
        self.extra_data = extra_data or {}

    def __repr__(self) -> str:
        """Return a representation of the remote repository information.

        Returns:
            str:
            The string representation.
        """
        return (
            f'<RemoteRepository: "{self.name}" (owner={self.owner}, '
            f'scm_type={self.scm_type})>'
        )



"""Change Group.mailing_list from an EmailField to a CharField.

This doesn't actually perform a field type change (not an option in Django
Evolution at this point), but it does set the attributes required to match an
`EmailField`.

Version Added:
    2.0
"""

from django_evolution.mutations import ChangeField


MUTATIONS = [
    ChangeField('Group', 'mailing_list', initial=None, max_length=254)
]



import kgb

from reviewboard.diffviewer.models import DiffSet, DiffSetHistory, FileDiff
from reviewboard.diffviewer.parser import DiffParser
from reviewboard.testing import TestCase


class DiffSetManagerTests(kgb.SpyAgency, TestCase):
    """Unit tests for DiffSetManager."""

    fixtures = ['test_scmtools']

    def test_create_from_data(self):
        """Testing DiffSetManager.create_from_data"""
        repository = self.create_repository(tool_name='Test')

        self.spy_on(repository.get_file_exists,
                    call_fake=lambda *args, **kwargs: True)

        diffset = DiffSet.objects.create_from_data(
            repository=repository,
            diff_file_name='diff',
            diff_file_contents=self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
            basedir='/')

        self.assertEqual(diffset.files.count(), 1)

    def test_create_from_data_with_basedir_no_slash(self):
        """Testing DiffSetManager.create_from_data with basedir without leading
        slash
        """
        repository = self.create_repository(tool_name='Test')

        self.spy_on(repository.get_file_exists,
                    call_fake=lambda *args, **kwargs: True)

        diffset = DiffSet.objects.create_from_data(
            repository=repository,
            diff_file_name='diff',
            diff_file_contents=self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
            basedir='trunk/')

        self.assertEqual(diffset.files.count(), 1)

        filediff = diffset.files.all()[0]
        self.assertEqual(filediff.source_file, 'trunk/README')
        self.assertEqual(filediff.dest_file, 'trunk/README')

    def test_create_from_data_with_basedir_slash(self):
        """Testing DiffSetManager.create_from_data with basedir with leading
        slash
        """
        repository = self.create_repository(tool_name='Test')

        self.spy_on(repository.get_file_exists,
                    call_fake=lambda *args, **kwargs: True)

        diffset = DiffSet.objects.create_from_data(
            repository=repository,
            diff_file_name='diff',
            diff_file_contents=self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
            basedir='/trunk/')

        self.assertEqual(diffset.files.count(), 1)

        filediff = diffset.files.all()[0]
        self.assertEqual(filediff.source_file, 'trunk/README')
        self.assertEqual(filediff.dest_file, 'trunk/README')

    def test_create_from_data_with_validate_only_true(self):
        """Testing DiffSetManager.create_from_data with validate_only=True"""
        repository = self.create_repository(tool_name='Test')

        self.spy_on(repository.get_file_exists,
                    call_fake=lambda *args, **kwargs: True)

        with self.assertNumQueries(0):
            diffset = DiffSet.objects.create_from_data(
                repository=repository,
                diff_file_name='diff',
                diff_file_contents=self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                basedir='/',
                validate_only=True)

        self.assertIsNone(diffset)
        self.assertEqual(DiffSet.objects.count(), 0)
        self.assertEqual(FileDiff.objects.count(), 0)

    def test_create_empty(self):
        """Testing DiffSetManager.create_empty"""
        repository = self.create_repository(tool_name='Test')
        history = DiffSetHistory.objects.create()

        diffset = DiffSet.objects.create_empty(
            repository=repository,
            diffset_history=history)

        self.assertEqual(diffset.files.count(), 0)
        self.assertEqual(diffset.revision, 1)

    def test_create_empty_with_revision(self):
        """Testing DiffSetManager.create_empty with revision"""
        repository = self.create_repository(tool_name='Test')
        history = DiffSetHistory.objects.create()

        diffset = DiffSet.objects.create_empty(
            repository=repository,
            diffset_history=history,
            revision=10)

        self.assertEqual(diffset.files.count(), 0)
        self.assertEqual(diffset.history, history)
        self.assertEqual(diffset.revision, 10)

    def test_create_empty_without_history(self):
        """Testing DiffSetManager.create_empty without diffset_history"""
        repository = self.create_repository(tool_name='Test')

        diffset = DiffSet.objects.create_empty(
            repository=repository)

        self.assertEqual(diffset.files.count(), 0)
        self.assertIsNone(diffset.history)
        self.assertEqual(diffset.revision, 0)

    def test_create_from_data_custom_parser_extra_data(self):
        """Testing DiffSetManager.create_from_data with a custom diff parser
        that sets extra_data
        """
        repository = self.create_repository(tool_name='Test')

        class CustomParser(DiffParser):
            def parse(self):
                result = super(CustomParser, self).parse()

                self.parsed_diff.extra_data = {
                    'key1': 'value1',
                }

                self.parsed_diff_change.extra_data = {
                    'key2': 'value2',
                }

                return result

            def parse_diff_header(self, linenum, parsed_file):
                parsed_file.extra_data = {
                    'key3': 'value3',
                }

                return super(CustomParser, self).parse_diff_header(
                    linenum, parsed_file)

        self.spy_on(repository.get_file_exists,
                    op=kgb.SpyOpReturn(True))

        tool = repository.get_scmtool()

        self.spy_on(repository.get_scmtool,
                    op=kgb.SpyOpReturn(tool))
        self.spy_on(tool.get_parser,
                    call_fake=lambda repo, diff: CustomParser(diff))

        diffset = DiffSet.objects.create_from_data(
            repository=repository,
            diff_file_name='diff',
            diff_file_contents=self.DEFAULT_FILEDIFF_DATA_DIFF,
            basedir='/')

        # Test against what's in the database.
        diffset.refresh_from_db()

        self.assertEqual(diffset.extra_data, {
            'change_extra_data': {
                'key2': 'value2',
            },
            'key1': 'value1',
        })

        self.assertEqual(diffset.files.count(), 1)

        filediff = diffset.files.all()[0]
        self.assertEqual(filediff.extra_data, {
            'is_symlink': False,
            'key3': 'value3',
            'raw_delete_count': 1,
            'raw_insert_count': 1,
        })

        # Check the FileLookupContext passed to get_file_exists.
        self.assertSpyCallCount(repository.get_file_exists, 1)

        context = repository.get_file_exists.last_call.kwargs.get('context')
        self.assertIsNotNone(context)
        self.assertEqual(context.diff_extra_data, {
            'key1': 'value1',
        })
        self.assertEqual(context.commit_extra_data, {
            'key2': 'value2',
        })
        self.assertEqual(context.file_extra_data, {
            'key3': 'value3',
        })

    def test_create_from_data_custom_parser_extra_data_and_parent(self):
        """Testing DiffSetManager.create_from_data with a custom diff parser
        that sets extra_data and using parent diff
        """
        repository = self.create_repository(tool_name='Test')

        namespaces = ['main_', 'parent_']

        class CustomParser(DiffParser):
            def parse(self):
                self.namespace = namespaces.pop(0)

                result = super(CustomParser, self).parse()

                self.parsed_diff.extra_data = {
                    '%skey1' % self.namespace: 'value1',
                }

                self.parsed_diff_change.extra_data = {
                    '%skey2' % self.namespace: 'value2',
                }

                return result

            def parse_diff_header(self, linenum, parsed_file):
                parsed_file.extra_data = {
                    '%skey3' % self.namespace: 'value3',
                }

                return super(CustomParser, self).parse_diff_header(
                    linenum, parsed_file)

        self.spy_on(repository.get_file_exists,
                    op=kgb.SpyOpReturn(True))

        tool = repository.get_scmtool()

        self.spy_on(repository.get_scmtool,
                    op=kgb.SpyOpReturn(tool))
        self.spy_on(tool.get_parser,
                    call_fake=lambda repo, diff: CustomParser(diff))

        diffset = DiffSet.objects.create_from_data(
            repository=repository,
            diff_file_name='diff',
            diff_file_contents=self.DEFAULT_FILEDIFF_DATA_DIFF,
            parent_diff_file_name='parent-diff',
            parent_diff_file_contents=self.DEFAULT_FILEDIFF_DATA_DIFF,
            basedir='/')

        # Test against what's in the database.
        diffset.refresh_from_db()

        self.assertEqual(diffset.extra_data, {
            'change_extra_data': {
                'main_key2': 'value2',
            },
            'main_key1': 'value1',
            'parent_extra_data': {
                'change_extra_data': {
                    'parent_key2': 'value2',
                },
                'parent_key1': 'value1',
            },
        })

        self.assertEqual(diffset.files.count(), 1)

        filediff = diffset.files.all()[0]
        self.assertEqual(filediff.extra_data, {
            '__parent_diff_empty': False,
            'is_symlink': False,
            'main_key3': 'value3',
            'parent_extra_data': {
                'parent_key3': 'value3',
            },
            'parent_source_filename': '/README',
            'parent_source_revision': 'revision 123',
            'raw_delete_count': 1,
            'raw_insert_count': 1,
        })

        # Check the FileLookupContext passed to get_file_exists.
        self.assertSpyCallCount(repository.get_file_exists, 1)

        context = repository.get_file_exists.last_call.kwargs.get('context')
        self.assertIsNotNone(context)
        self.assertEqual(context.diff_extra_data, {
            'parent_key1': 'value1',
        })
        self.assertEqual(context.commit_extra_data, {
            'parent_key2': 'value2',
        })
        self.assertEqual(context.file_extra_data, {
            'parent_key3': 'value3',
        })



"""Search indexing for review requests."""

from __future__ import annotations

from typing import TYPE_CHECKING, cast

from django.contrib.auth.models import AnonymousUser, User
from django.db.models import Prefetch
from haystack import indexes

from reviewboard.diffviewer.models import DiffSet, FileDiff
from reviewboard.reviews.models import Group, ReviewRequest
from reviewboard.scmtools.models import Repository
from reviewboard.search.indexes import BaseSearchIndex
from reviewboard.site.models import LocalSite

if TYPE_CHECKING:
    from collections.abc import Sequence
    from reviewboard.accounts.models import User as RBUser

    from django.db.models import QuerySet


class ReviewRequestIndex(BaseSearchIndex[ReviewRequest],
                         indexes.Indexable):
    """A Haystack search index for Review Requests."""

    model = ReviewRequest
    local_site_attr = 'local_site_id'

    # We shouldn't use 'id' as a field name because it's by default reserved
    # for Haystack. Hiding it will cause duplicates when updating the index.
    review_request_id = indexes.IntegerField(model_attr='display_id')
    summary = indexes.CharField(model_attr='summary')
    description = indexes.CharField(model_attr='description')
    testing_done = indexes.CharField(model_attr='testing_done')
    commit_id = indexes.EdgeNgramField(model_attr='commit', null=True)
    bug = indexes.CharField(model_attr='bugs_closed')
    username = indexes.CharField(model_attr='submitter__username')
    author = indexes.CharField()
    last_updated = indexes.DateTimeField(model_attr='last_updated')
    url = indexes.CharField(model_attr='get_absolute_url')
    file = indexes.CharField()

    # These fields all contain information needed to perform queries about
    # whether a review request is accessible by a given user.
    private = indexes.BooleanField()
    private_repository_id = indexes.IntegerField()
    private_target_groups = indexes.MultiValueField()
    target_users = indexes.MultiValueField()

    def get_updated_field(self) -> str:
        """Return the name of the field indicating the last updated timestamp.

        Returns:
            str:
            The name of the field.
        """
        return 'last_updated'

    def index_queryset(
        self,
        using: (str | None) = None,
    ) -> QuerySet[ReviewRequest]:
        """Return a queryset for indexing review requests.

        This will only include review requests that are public.

        Args:
            using (str, unused):
                The name of the search index.

        Returns:
            django.db.models.QuerySet:
            The queryset to index.
        """
        has_local_sites = LocalSite.objects.has_local_sites()
        extra_prefetches: list[Prefetch] = []
        local_site_acl_fields: list[str] = []
        local_site_select_related: list[str | None] = []

        if has_local_sites:
            # If the server has Local Sites, prefetch them for all review
            # requests and select_related them for other relevant queries.
            #
            # If a Local Site is added for the first time while indexing,
            # this will result in N+1 query issues for every review request
            # or related object, which will be expensive. This would only
            # ever happen once, and should be an unlikely event.
            extra_prefetches.append(
                Prefetch(
                    'local_site',
                    queryset=(
                        LocalSite.objects
                        .only(
                            'name',
                        )
                    ),
                ),
            )

            local_site_acl_fields.append('local_site__public')
            local_site_select_related.append('local_site')
        else:
            local_site_select_related.append(None)

        return (
            self.get_model().objects
            .public(
                status=None,
                local_site=LocalSite.ALL,
                show_inactive=True,
                filter_private=False,
                distinct=False,
            )
            .only(
                # ReviewRequest fields.
                'bugs_closed',
                'changenum',
                'commit_id',
                'description',
                'last_updated',
                'local_id',
                'local_site_id',
                'public',
                'repository_id',
                'submitter_id',
                'summary',
                'testing_done',

                # DiffSetHistory fields.
                'diffset_history__id',
            )
            .select_related(
                'diffset_history',
            )
            .prefetch_related(
                Prefetch(
                    'repository',
                    queryset=(
                        Repository.objects
                        .only(
                            'local_site',
                            'public',

                            *local_site_acl_fields,
                        )
                        .select_related(*local_site_select_related)
                    )
                ),
                Prefetch(
                    'submitter',
                    queryset=(
                        User.objects
                        .select_related(
                            'profile',
                        )
                        .only(
                            # User fields.
                            'first_name',
                            'last_name',
                            'username',

                            # Profile fields.
                            'profile__extra_data',
                            'profile__is_private',
                        )
                    ),
                ),

                # Note that we're handling this nested relation very
                # carefully. A standard prefetch will fetch all columns for
                # each object in the relation specifier, while we really only
                # need the IDs necessary to chain the objects together.
                Prefetch(
                    'diffset_history__diffsets',
                    queryset=(
                        DiffSet.objects
                        .only(
                            'history_id',
                        )
                        .prefetch_related(
                            Prefetch(
                                'files',
                                queryset=(
                                    FileDiff.objects
                                    .only(
                                        'dest_file',
                                        'source_file',
                                    )
                                ),
                            ),
                        )
                    ),
                ),
                Prefetch(
                    'target_groups',
                    queryset=(
                        Group.objects
                        .only(
                            # Group fields.
                            'invite_only',
                            'local_site',

                            *local_site_acl_fields,
                        )
                        .select_related(*local_site_select_related)
                    ),
                ),
                Prefetch(
                    'target_people',
                    queryset=(
                        User.objects
                        .only('pk')
                    ),
                ),
                *extra_prefetches,
            )
        )

    def prepare_file(
        self,
        review_request: ReviewRequest,
    ) -> set[tuple[str, str]]:
        """Prepare the diff file information for the index.

        This will cover all filediffs across diffsets.

        Args:
            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request being prepared.

        Returns:
            set:
            A set of tuples of:

            Tuple:
                0 (str):
                    The source file path.

                1 (str):
                    The destination file path.
        """
        return {
            (filediff.source_file, filediff.dest_file)
            for diffset in review_request.diffset_history.diffsets.all()
            for filediff in diffset.files.all()
        }

    def prepare_private(
        self,
        review_request: ReviewRequest,
    ) -> bool:
        """Prepare the private flag for the index.

        This will be set to true if the review request isn't generally
        accessible to users.

        Args:
            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request being prepared.

        Returns:
            bool:
            ``True`` if the review request is private. ``False`` if public.
        """
        return not review_request.is_accessible_by(AnonymousUser(),
                                                   silent=True)

    def prepare_private_repository_id(
        self,
        review_request: ReviewRequest,
    ) -> int:
        """Prepare the private repository ID, if any, for the index.

        If there's no repository, or it's public, 0 will be returned instead.

        Args:
            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request being prepared.

        Returns:
            int:
            The ID of the repository, or 0 if unset or public.
        """
        if review_request.repository and not review_request.repository.public:
            return review_request.repository_id
        else:
            return 0

    def prepare_private_target_groups(
        self,
        review_request: ReviewRequest,
    ) -> Sequence[int]:
        """Prepare the list of invite-only target groups for the index.

        If there aren't any invite-only groups associated, ``[0]`` will be
        returned. This allows queries to be performed that check that none
        of the groups are private, since we can't query against empty lists.

        Args:
            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request being prepared.

        Returns:
            list of int:
            The list of invite-only group IDs, or ``[0]`` if there aren't
            any.
        """
        return [
            group.pk
            for group in review_request.target_groups.all()
            if group.invite_only
        ] or [0]

    def prepare_target_users(
        self,
        review_request: ReviewRequest,
    ) -> Sequence[int]:
        """Prepare the list of target users for the index.

        If there aren't any target users, ``[0]`` will be returned. This
        allows queries to be performed that check that there aren't any
        users in the list, since we can't query against empty lists.

        Args:
            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request being prepared.

        Returns:
            list of int:
            The list of user IDs, or ``[0]`` if there aren't any.
        """
        return [
            user.pk
            for user in review_request.target_people.all()
        ] or [0]

    def prepare_author(
        self,
        review_request: ReviewRequest,
    ) -> str:
        """Prepare the author field.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request being indexed.

        Returns:
            str:
            Either the author's full name (if their profile is public) or an
            empty string.
        """
        user = cast('RBUser', review_request.submitter)
        profile = user.get_profile(cached_only=True)

        if profile is None or profile.is_private:
            return ''

        return user.get_full_name()



"""Providers for managing licensing options.

Version Added:
    8.0
"""

from __future__ import annotations

import logging
from datetime import datetime, timedelta
from typing import Final, Generic, TYPE_CHECKING, cast
from uuid import uuid4

from django.utils.html import format_html
from django.utils.translation import gettext as _
from typelets.django.json import SerializableDjangoJSONDict
from typelets.json import JSONValue
from typing_extensions import TypeVar

from reviewboard.licensing.errors import LicenseActionError
from reviewboard.licensing.license import LicenseInfo, LicenseStatus

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import ClassVar

    from django.http import HttpRequest
    from typelets.json import JSONDictImmutable
    from typelets.django.json import SerializableDjangoJSONList

    from reviewboard.licensing.actions import (LicenseAction,
                                               LicenseActionData,
                                               LicenseActionHandler)
    from reviewboard.licensing.license import LicenseLineItem
    from reviewboard.licensing.license_checks import (
        RequestCheckLicenseResult,
        ProcessCheckLicenseResult,
    )


logger = logging.getLogger(__name__)


#: A type variable for license information classes.
#:
#: Version Added:
#:    8.0
_TLicenseInfo = TypeVar('_TLicenseInfo',
                        bound=LicenseInfo,
                        default=LicenseInfo)


class BaseLicenseProvider(Generic[_TLicenseInfo]):
    """Base class for a provider for managing licenses.

    License providers make internal licensing information available to
    Review Board, to simplify and unify license management for multiple
    add-ons in one place.

    A license provider can control what licenses they want to expose for
    display purposes, handle checking for changes to licenses (new, updated,
    or removed), and can provide custom actions that can be invoked via an
    action button.

    Version Added:
        8.0
    """

    #: A set of built-in actions common to all license providers.
    BUILTIN_ACTIONS: Final[set[str]] = {
        'license-update-check',
        'process-license-update',
        'upload-license',
    }

    #: The unique ID of the license provider.
    license_provider_id: ClassVar[str]

    #: The name of the JavaScript model for the license front-end.
    js_license_model_name: ClassVar[str] = 'RB.License'

    #: The name of the JavaScript view for the license front-end.
    js_license_view_name: ClassVar[str] = 'RB.LicenseView'

    #: Custom actions that can be handled by the license provider.
    #:
    #: Each action must correspond to a :samp:`handle_{actionname}_action`
    #: method.
    custom_actions: set[str] = set()

    def get_licenses(self) -> Sequence[_TLicenseInfo]:
        """Return the list of available licenses.

        This may include active licenses, expired licenses, and trial
        licenses.

        Returns:
            list of reviewboard.licensing.license.LicenseInfo:
            The list of licenses provided.
        """
        raise NotImplementedError

    def get_license_by_id(
        self,
        license_id: str,
    ) -> _TLicenseInfo | None:
        """Return a license with the given provider-unique ID.

        This may return an active license, expired license, or trial license.

        Args:
            license_id (str):
                The provider-unique ID for the license.

        Returns:
            reviewboard.licensing.license.LicenseInfo:
            The license information, or ``None`` if not found.
        """
        raise NotImplementedError

    def get_manage_license_url(
        self,
        *,
        license_info: _TLicenseInfo,
    ) -> str | None:
        """Return the URL for managing licenses online.

        This is used to link to an outside license management portal for a
        given license.

        Args:
            license_info (reviewboard.licensing.license.LicenseInfo):
                The license information to link to in the portal.

        Returns:
            str:
            The URL to the license portal, or ``None`` if one is not
            available.
        """
        return None

    def get_license_actions(
        self,
        *,
        license_info: _TLicenseInfo,
        request: (HttpRequest | None) = None,
    ) -> list[LicenseAction]:
        """Return actions to display when viewing a license.

        By default, this may contain Manage License and Update License
        actions. Subclasses may override this to provide custom actions.

        Args:
            license_info (reviewboard.licensing.license.LicenseInfo):
                The license information that will be displayed.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

        Returns:
            list of reviewboard.licensing.actions.LicenseAction:
            The list of actions to display for the license.
        """
        actions: list[LicenseAction] = []

        manage_url = self.get_manage_license_url(license_info=license_info)

        if manage_url:
            actions.append({
                'action_id': 'manage-license',
                'label': _('Manage your license'),
                'primary': True,
                'url': manage_url,
            })

        if license_info.can_upload_license:
            actions.append({
                'action_id': 'upload-license',
                'label': _('Upload a new license file'),
            })

        return actions

    def get_license_line_items(
        self,
        *,
        license_info: _TLicenseInfo,
        request: (HttpRequest | None) = None,
    ) -> Sequence[LicenseLineItem]:
        """Return line items to display when viewing a license.

        Line items may contain information or even custom HTML describing
        an important aspect of a license.

        By default, this will contain any line items assigned to the license.
        Subclasses may override this to provide additional information.

        Args:
            license_info (reviewboard.licensing.license.LicenseInfo):
                The license information that will be displayed.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

        Returns:
            list of reviewboard.licensing.license.LicenseLineItem:
            The list of line items to display for the license.
        """
        return license_info.line_items

    def get_check_license_request(
        self,
        *,
        action_data: JSONDictImmutable,
        license_info: _TLicenseInfo,
        request: (HttpRequest | None) = None,
    ) -> RequestCheckLicenseResult | None:
        """Return data used to check a license for validity.

        This can be any JSON data that the license provider may want to
        send to a license portal in order to check if a license is still
        valid or if there's a newer license available.

        Args:
            action_data (dict):
                Data provided in the request to the action.

                This will correspond to HTTP POST data if processing via an
                HTTP request from the client.

            license_info (reviewboard.licensing.license.LicenseInfo):
                The license information that will be checked.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client, if any.

        Returns:
            reviewboard.licensing.license_checks.RequestCheckLicenseResult:
            The state used to control the license check, or ``None`` if checks
            are not supported.

        Raises:
            reviewboard.licensing.errors.LicenseActionError:
                There was an error generating data for this request.
        """
        return None

    def get_js_license_model_data(
        self,
        *,
        license_info: _TLicenseInfo,
        request: (HttpRequest | None) = None,
    ) -> SerializableDjangoJSONDict:
        """Return data for the JavaScript license model.

        This provides all the data needed by the JavaScript license model
        to display and manage the license in the UI.

        Args:
            license_info (reviewboard.licensing.license.LicenseInfo):
                The license information to convert to model data.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

        Returns:
            typelets.django.json.SerializableDjangoJSONDict:
            The data for the JavaScript license model.
        """
        is_trial = license_info.is_trial
        plan_name = license_info.plan_name
        product = license_info.product_name
        status = license_info.status

        # Calculate expiration information.
        expires = license_info.expires
        grace_period_days = license_info.grace_period_days_remaining
        hard_expires_date: datetime | None

        if expires and grace_period_days:
            hard_expires_date = expires + timedelta(days=grace_period_days)
        else:
            hard_expires_date = expires

        # Generate a notice, if needed.
        notice_html: str = ''

        if status == LicenseStatus.EXPIRED_GRACE_PERIOD:
            # NOTE: This is in Ink syntax, not raw HTML.
            datetime_ink = format_html(
                '<time class="timesince" dateTime="{expires_date}"/>',
                expires_date=hard_expires_date)

            notice_html = format_html(
                _('Your grace period is now active. Unless renewed, '
                  '{product} will be disabled {datetime_ink}.'),
                datetime_ink=datetime_ink,
                product=product)

        # Build actions for the license.
        actions_data: SerializableDjangoJSONList = [
            {
                'actionID': action['action_id'],
                'label': action['label'],
                **{
                    result_key: cast(JSONValue, action.get(action_key))
                    for result_key, action_key in (
                        ('callArgs', 'call_args'),
                        ('extraData', 'extra_data'),
                        ('primary', 'primary'),
                        ('url', 'url')
                    )
                    if action.get(action_key)
                },
            }
            for action in self.get_license_actions(license_info=license_info)
        ]

        # Build line items for the license.
        line_items_data: SerializableJSONList = [
            {
                'content': line_item['content'],
                **{
                    result_key: cast(JSONValue, line_item.get(line_item_key))
                    for result_key, line_item_key in (
                        ('contentIsHTML', 'content_is_html'),
                        ('icon', 'icon'),
                    )
                    if line_item.get(line_item_key)
                },
            }
            for line_item in self.get_license_line_items(
                license_info=license_info)
        ]

        return {
            'actionTarget': (
                f'{self.license_provider_id}:{license_info.license_id}'
            ),
            'actions': actions_data,
            'autoRenew': license_info.auto_renew,
            'canUploadLicense': license_info.can_upload_license,
            'expiresDate': expires,
            'expiresSoon': license_info.get_expires_soon(),
            'gracePeriodDaysRemaining': grace_period_days,
            'hardExpiresDate': hard_expires_date,
            'isTrial': is_trial,
            'licenseID': license_info.license_id,
            'licensedTo': license_info.licensed_to,
            'lineItems': line_items_data,
            'manageURL': self.get_manage_license_url(
                license_info=license_info),
            'noticeHTML': notice_html,
            'planID': license_info.plan_id,
            'planName': plan_name,
            'productName': product,
            'status': status.value,
            'summary': license_info.get_summary(),
        }

    def process_check_license_result(
        self,
        *,
        action_data: LicenseActionData,
        license_info: _TLicenseInfo,
        check_request_data: JSONValue,
        check_response_data: JSONValue,
        session_token: str | None,
        request: (HttpRequest | None) = None,
    ) -> ProcessCheckLicenseResult:
        """Process the result of a license check.

        This will take the response from a license check and process it. This
        may involve applying and saving the new license, or updating related
        state.

        The result indicates if the license is already up-to-date,
        newly-applied, or hit an error. It may also return new license
        attributes.

        Args:
            action_data (reviewboard.licensing.actions.LicenseActionData):
                Data provided in the request to the action.

                This will correspond to HTTP POST data if processing via an
                HTTP request from the client.

            license_info (reviewboard.licensing.license.LicenseInfo):
                The license information to convert to model data.

            check_request_data (typelets.json.JSONValue):
                The original request data sent to the license check.

            check_response_data (typelets.json.JSONValue):
                The response data from the license check.

            session_token (str):
                The session token provided by
                :py:meth:`get_check_license_request`.

                This will be ``None`` if a session token was not provided.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

        Returns:
            reviewboard.licensing.license_checks.ProcessCheckLicenseResult:
            The processed result of the license check.

        Raises:
            NotImplementedError:
                This method must be implemented by subclasses.

            reviewboard.licensing.errors.LicenseActionError:
                There was an error processing license data, or the license
                data was not valid for the product.
        """
        raise NotImplementedError

    def set_license_data(
        self,
        *,
        license_info: _TLicenseInfo,
        license_data: bytes,
        request: (HttpRequest | None) = None,
    ) -> None:
        """Manually set new license data.

        This is used to allow a user to upload a new license directly,
        if supported by the license provider.

        Subclasses that override this must validate the license and ensure
        it can be set.

        Args:
            license_info (reviewboard.licensing.license.LicenseInfo):
                The license information to convert to model data.

            license_data (bytes):
                The license data to set.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

        Raises:
            reviewboard.licensing.errors.LicenseActionError:
                There was an error setting license data, or the license
                data was not valid for the product.
        """
        raise NotImplementedError

    def call_action(
        self,
        action_id: str,
        *,
        license_info: _TLicenseInfo,
        action_data: LicenseActionData = {},
        request: (HttpRequest | None) = None,
    ) -> SerializableDjangoJSONDict:
        """Call and handle a registered license action.

        This is used to perform license action requests from the UI or within
        the server for purposes such as license update checks and activation.

        Custom handlers can include the following standard fields in a JSON
        response:

        Keys:
            license_infos (list of dict, optional):
                A mapping of license IDs to attributes. Any new licenses will
                be added to the display. Any licenses set to ``None`` will
                be removed from the display. Anything else will be updated.

            redirect_url (str, optional):
                A URL to redirect on the client.

        Args:
            action_data (dict):
                Data provided in the request to the action.

                This will correspond to HTTP POST data if processing via an
                HTTP request from the client.

            license_info (reviewboard.licensing.license.LicenseInfo):
                Information on the license to check for updates.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

        Returns:
            dict:
            The JSON-serializable dictionary of results to send back to the
            client.

        Raises:
            reviewboard.licensing.errors.LicenseActionError:
                An error invoking an action in the License Provider. This
                will result in a suitable error message for the client.
        """
        norm_action_id = action_id.replace('-', '_')
        handler_name = f'handle_{norm_action_id}_action'

        handler: (LicenseActionHandler | None) = None

        if (action_id in self.BUILTIN_ACTIONS or
            action_id in self.custom_actions):
            try:
                handler = getattr(self, handler_name)
            except AttributeError:
                pass

        if handler is None:
            raise LicenseActionError(
                _('Unsupported license action "{action}".')
                .format(action=action_id))

        return handler(license_info=license_info,
                       action_data=action_data,
                       request=request)

    def handle_license_update_check_action(
        self,
        *,
        action_data: LicenseActionData,
        license_info: _TLicenseInfo,
        request: HttpRequest | None,
    ) -> SerializableDjangoJSONDict:
        """Handle a license update check.

        This will request a license server URL and a payload from the License
        Provider. The client can send the payload to the license server URL to
        request a new license, or check status for a license.

        Args:
            action_data (dict):
                Data provided in the request to the action.

                This will correspond to HTTP POST data if processing via an
                HTTP request from the client.

            license_info (reviewboard.licensing.license.LicenseInfo):
                Information on the license to check for updates.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

        Returns:
            dict:
            The JSON-serializable dictionary of results to send back to the
            client.

        Raises:
            reviewboard.licensing.errors.LicenseActionError:
                An error invoking an action in the License Provider. This
                will result in a suitable error message for the client.
        """
        check_request = self.get_check_license_request(
            license_info=license_info,
            action_data=action_data,
            request=request)

        if not check_request:
            return {
                'canCheck': False,
            }

        return {
            'canCheck': True,
            'checkStatusURL': check_request['url'],
            'credentials': check_request.get('credentials'),
            'data': check_request['data'],
            'headers': check_request.get('headers'),
            'sessionToken': check_request.get('session_token'),
        }

    def handle_process_license_update_action(
        self,
        *,
        action_data: LicenseActionData,
        license_info: _TLicenseInfo,
        request: HttpRequest | None,
    ) -> SerializableDjangoJSONDict:
        """Handle an automated license update payload.

        This will process the payload from a license server, passing the
        result to the License Provider. That may install a new license or
        update information about a license in the backend.

        Args:
            action_data (dict):
                Data provided in the request to the action.

                This will correspond to HTTP POST data if processing via an
                HTTP request from the client.

            license_info (reviewboard.licensing.license.LicenseInfo):
                Information on the license to check for updates.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

        Returns:
            dict:
            The JSON-serializable dictionary of results to send back to the
            client.

        Raises:
            reviewboard.licensing.errors.LicenseActionError:
                An error with the data or with invoking the actions in the
                License Provider. This will result in a suitable error
                message for the client.
        """
        trace_id = str(uuid4())

        logger.info('[%s] Checking license update response for License '
                    'Provider %r',
                    trace_id, self.license_provider_id)

        # Check the session token, if provided.
        session_token = action_data.get('session_token')

        if session_token is not None:
            logger.debug('[%s] Session token = %r',
                         trace_id, session_token)

            if not isinstance(session_token, str):
                raise LicenseActionError(
                    _(
                        'Unsupported session_token type for license check. '
                        'This may be an internal error or an issue with the '
                        'licensing server. Check the Review Board server logs '
                        'for more information (error ID {trace_id}).'
                    ).format(trace_id=trace_id))

        # Pull out the request and response data to verify.
        try:
            check_request_data = action_data['check_request_data']
        except KeyError:
            logger.error('[%s] Missing check_request_data for license check',
                         trace_id)

            raise LicenseActionError(
                _(
                    'Missing check_request_data value for license check. '
                    'This may be an internal error or an issue with the '
                    'licensing server. Check the Review Board server logs '
                    'for more information (error ID {trace_id}).'
                ).format(trace_id=trace_id))

        logger.debug('[%s] Check request data = %r',
                     trace_id, check_request_data)

        try:
            check_response_data = action_data['check_response_data']
        except KeyError:
            logger.error('[%s] Missing check_response_data for license check',
                         trace_id)

            raise LicenseActionError(
                _(
                    'Missing check_response_data value for license check. '
                    'This may be an internal error or an issue with the '
                    'licensing server. Check the Review Board server logs '
                    'for more information (error ID {trace_id}).'
                ).format(trace_id=trace_id))

        logger.debug('[%s] Check response data = %r',
                     trace_id, check_response_data)

        if not check_response_data:
            logger.error('[%s] Empty check_response_data for license check',
                         trace_id)

            raise LicenseActionError(
                _(
                    'Empty check_response_data value for license check. '
                    'This may be an internal error or an issue with the '
                    'licensing server. Check the Review Board server logs '
                    'for more information (error ID {trace_id}).'
                ).format(trace_id=trace_id))

        # Pass to the License Provider and check the result.
        try:
            result = self.process_check_license_result(
                action_data=action_data,
                license_info=license_info,
                check_request_data=check_request_data,
                check_response_data=check_response_data,
                session_token=session_token,
                request=request)
            status = result['status']
        except NotImplementedError as e:
            logger.exception('[%s] Automated license checks are enabled '
                             'for this license provider but not implemented. '
                             'This is an internal error.',
                             trace_id)

            raise LicenseActionError(
                _(
                    'The license provider implementation enables support for '
                    'automated license checks but does not provide an '
                    'implementation. This is an internal error. See the '
                    'Review Board server logs for more information (error ID '
                    '{trace_id}).'
                ).format(trace_id=trace_id)
            ) from e
        except LicenseActionError as e:
            raise LicenseActionError(
                _(
                    'Error processing license update: {message}'
                ).format(message=str(e)),
                payload=e.payload,
            ) from e
        except Exception as e:
            logger.exception('[%s] Unexpected error checking license '
                             'result: %s',
                             trace_id, e)

            raise LicenseActionError(
                _(
                    'Unexpected error processing license update. Check the '
                    'Review Board server logs for more information (error ID '
                    '{trace_id}).'
                ).format(trace_id=trace_id)
            ) from e

        logger.info('[%s] License update check complete: %s',
                    trace_id, status)

        return cast(SerializableDjangoJSONDict, result)

    def handle_upload_license_action(
        self,
        *,
        action_data: LicenseActionData,
        license_info: _TLicenseInfo,
        request: HttpRequest | None,
    ) -> SerializableDjangoJSONDict:
        """Handle a manual license upload.

        This will take the provided license data and pass it to the License
        Provider for license replacement. This is dependent on the
        capabilities of the License Provider.

        Args:
            action_data (dict):
                Data provided in the request to the action.

                This will correspond to HTTP POST data if processing via an
                HTTP request from the client.

            license_info (reviewboard.licensing.license.LicenseInfo):
                Information on the license to check for updates.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

        Returns:
            dict:
            The JSON-serializable dictionary of results to send back to the
            client.

        Raises:
            reviewboard.licensing.errors.LicenseActionError:
                An error with the data or with invoking the upload in the
                License Provider. This will result in a suitable error
                message for the client.
        """
        license_data: (bytes | None) = None

        if request:
            license_data_fp = request.FILES.get('license_data')

            if license_data_fp is not None:
                license_data = license_data_fp.read()

        if not license_data:
            raise LicenseActionError(_('No license data was found'))

        # Allow any errors to bubble up.
        try:
            self.set_license_data(license_info=license_info,
                                  license_data=license_data,
                                  request=request)
        except NotImplementedError:
            raise LicenseActionError(_(
                'Licenses for this product cannot be uploaded manually.'
            ))
        except LicenseActionError:
            # Let this error bubble up.
            raise
        except Exception as e:
            logger.exception('Unexpected error setting license data %r for '
                             'license %r on provider %r: %s',
                             license_data, license_info, self, e)

            raise LicenseActionError(_(
                'Unexpected error setting license data for this product. '
                'Check the Review Board server logs for more information.'
            ))

        # Reload the license.
        new_license_info = self.get_license_by_id(license_info.license_id)

        if new_license_info:
            license_info_data = self.get_js_license_model_data(
                license_info=new_license_info,
                request=request)
        else:
            license_info_data = None

        return {
            'license_info': license_info_data,
        }
